"""implementation details for building a bioimage.io resource description"""

import collections.abc
from typing import Any, Callable, List, Literal, Mapping, Optional, Type, TypeVar, Union

from ._internal.common_nodes import InvalidDescr, ResourceDescrBase
from ._internal.field_validation import issue_warning
from ._internal.io import BioimageioYamlContentView
from ._internal.types import FormatVersionPlaceholder
from ._internal.validation_context import ValidationContext, get_validation_context
from .summary import (
    ErrorEntry,
    ValidationDetail,
)

ResourceDescrT = TypeVar("ResourceDescrT", bound=ResourceDescrBase)


DISCOVER: Literal["discover"] = "discover"
"""placeholder for whatever format version an RDF specifies"""


def get_rd_class_impl(
    typ: Any,
    format_version: Any,
    descriptions_map: Mapping[Optional[str], Mapping[str, Type[ResourceDescrT]]],
    fallback_to_latest: bool,
) -> Type[ResourceDescrT]:
    """get the resource description class for the given type and format version"""
    assert None in descriptions_map
    assert all("latest" in version_map for version_map in descriptions_map.values())
    assert all(
        fv == "latest" or fv.count(".") == 1
        for version_map in descriptions_map.values()
        for fv in version_map
    )
    if not isinstance(typ, str) or typ not in descriptions_map:
        typ = None

    format_version = str(format_version)
    if format_version == "latest" or (ndots := format_version.count(".")) == 1:
        use_format_version = format_version
    elif ndots == 0:
        use_format_version = format_version + ".0"
    else:
        assert ndots > 1
        use_format_version = format_version[: format_version.rfind(".")]

    descr_versions = descriptions_map[typ]
    if use_format_version not in descr_versions:
        if fallback_to_latest:
            issue_warning(
                "Unsupported format version '{value}' for type '{typ}'.",
                value=format_version,
                field="format_version",
                log_depth=3,
                msg_context={"typ": typ},
            )
            use_format_version = "latest"
        else:
            raise ValueError(
                f"Unsupported format version '{format_version}' for type '{typ}'."
                + " Supported format versions are: {', '.join(sorted(fv for fv in descr_versions))}"
            )

    return descr_versions[use_format_version]


def build_description_impl(
    content: BioimageioYamlContentView,
    /,
    *,
    context: Optional[ValidationContext] = None,
    format_version: Union[FormatVersionPlaceholder, str] = DISCOVER,
    get_rd_class: Callable[[Any, Any, bool], Type[ResourceDescrT]],
) -> Union[ResourceDescrT, InvalidDescr]:
    context = context or get_validation_context()
    errors: List[ErrorEntry] = []
    if isinstance(content, collections.abc.Mapping):
        for minimum in ("type", "format_version"):
            if minimum not in content:
                errors.append(
                    ErrorEntry(
                        loc=(minimum,), msg=f"Missing field '{minimum}'", type="error"
                    )
                )
            elif not isinstance(content[minimum], str):
                errors.append(
                    ErrorEntry(
                        loc=(minimum,),
                        msg=f"Invalid type '{type(content[minimum])}'",
                        type="error",
                    )
                )
    else:
        errors.append(
            ErrorEntry(
                loc=(), msg=f"Invalid content of type '{type(content)}'", type="error"
            )
        )
        content = {}

    if errors:
        ret = InvalidDescr(**content)  # pyright: ignore[reportArgumentType]
        ret.validation_summary.add_detail(
            ValidationDetail(
                name="extract fields to chose description class",
                status="failed",
                errors=errors,
                context=context.summary,
            )
        )
        return ret

    typ = content["type"]
    # check format_version argument before loading as 'discover'
    # to throw an exception for an invalid format_version early
    if str(format_version).lower() != DISCOVER:
        as_rd_class = get_rd_class(typ, format_version, False)
    else:
        as_rd_class = None
    # always load with discovered format_version first
    rd_class = get_rd_class(typ, content["format_version"], True)
    rd = rd_class.load(content, context=context)

    if as_rd_class is not None and as_rd_class is not rd_class:
        # load with requested format_version
        discover_details = rd.validation_summary.details
        rd = as_rd_class.load(content, context=context)
        assert rd.validation_summary is not None
        rd.validation_summary.details[:0] = discover_details

    return rd



"""Utility functions for bioimage.io specifications (mostly IO)."""

import json
from typing import Any, Dict, List, TypedDict, Union

from imageio.v3 import imread  # pyright: ignore[reportUnknownVariableType]
from numpy.typing import NDArray

from ._description import ensure_description_is_dataset as ensure_description_is_dataset
from ._description import ensure_description_is_model as ensure_description_is_model
from ._internal.io import FileDescr
from ._internal.io import download as download
from ._internal.io import extract_file_name as extract_file_name
from ._internal.io import get_reader as get_reader
from ._internal.io import get_sha256 as get_sha256
from ._internal.io import (
    identify_bioimageio_yaml_file_name as identify_bioimageio_yaml_file_name,
)
from ._internal.io import interprete_file_source as interprete_file_source
from ._internal.io import is_valid_bioimageio_yaml_name as is_valid_bioimageio_yaml_name
from ._internal.io_basics import ZipPath
from ._internal.io_utils import load_array as load_array
from ._internal.io_utils import open_bioimageio_yaml as open_bioimageio_yaml
from ._internal.io_utils import read_yaml as read_yaml
from ._internal.io_utils import save_array as save_array
from ._internal.io_utils import write_yaml as write_yaml
from ._internal.type_guards import is_ndarray
from ._internal.types import PermissiveFileSource, RelativeFilePath
from ._internal.utils import files

get_file_name = extract_file_name


class SpdxLicenseEntry(TypedDict):
    isDeprecatedLicenseId: bool
    isKnownByZenodo: bool
    isOsiApproved: bool
    licenseId: str
    name: str
    reference: str


class SpdxLicenses(TypedDict):
    licenseListVersion: str
    licenses: List[SpdxLicenseEntry]
    releaseDate: str


def get_spdx_licenses() -> SpdxLicenses:
    """get details of the SPDX licenses known to bioimageio.spec"""
    with (
        files("bioimageio.spec")
        .joinpath("static/spdx_licenses.json")
        .open("r", encoding="utf-8")
    ) as f:
        return json.load(f)


def get_bioimageio_json_schema() -> Dict[str, Any]:
    """get the bioimageio specification as a JSON schema"""
    with (
        files("bioimageio.spec")
        .joinpath("static/bioimageio_schema.json")
        .open("r", encoding="utf-8")
    ) as f:
        return json.load(f)


def load_image(source: Union[FileDescr, ZipPath, PermissiveFileSource]) -> NDArray[Any]:
    """load a single image as numpy array

    Args:
        source: image source
    """

    if isinstance(source, (FileDescr, ZipPath)):
        parsed_source = source
    else:
        parsed_source = interprete_file_source(source)

    if isinstance(parsed_source, RelativeFilePath):
        parsed_source = parsed_source.absolute()

    if parsed_source.suffix == ".npy":
        image = load_array(parsed_source)
    else:
        reader = get_reader(parsed_source)
        image = imread(  # pyright: ignore[reportUnknownVariableType]
            reader.read(), extension=parsed_source.suffix
        )

    assert is_ndarray(image)
    return image



from typing import Any, ClassVar, Type

from pydantic import RootModel

from ._generated_spdx_license_literals import (
    DeprecatedLicenseId as DeprecatedLicenseIdLiteral,
)
from ._generated_spdx_license_literals import LicenseId as LicenseIdLiteral
from .validated_string import ValidatedString

__all__ = [
    "DeprecatedLicenseId",
    "DeprecatedLicenseIdLiteral",
    "LicenseId",
    "LicenseIdLiteral",
]


class LicenseId(ValidatedString):
    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[LicenseIdLiteral]


class DeprecatedLicenseId(ValidatedString):
    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[DeprecatedLicenseIdLiteral]



from typing import TYPE_CHECKING, ClassVar, Literal, Optional, Union

from typing_extensions import Annotated

from .._internal.common_nodes import Node
from .._internal.io import WithSuffix
from .._internal.types import FilePath
from .._internal.url import HttpUrl
from ..generic.v0_2 import VALID_COVER_IMAGE_EXTENSIONS as VALID_COVER_IMAGE_EXTENSIONS
from ..generic.v0_2 import AttachmentsDescr as AttachmentsDescr
from ..generic.v0_2 import Author as Author
from ..generic.v0_2 import BadgeDescr as BadgeDescr
from ..generic.v0_2 import CiteEntry as CiteEntry
from ..generic.v0_2 import Doi as Doi
from ..generic.v0_2 import GenericDescrBase
from ..generic.v0_2 import LinkedResource as LinkedResource
from ..generic.v0_2 import Maintainer as Maintainer
from ..generic.v0_2 import OrcidId as OrcidId
from ..generic.v0_2 import RelativeFilePath as RelativeFilePath
from ..generic.v0_2 import ResourceId as ResourceId
from ..generic.v0_2 import Uploader as Uploader
from ..generic.v0_2 import Version as Version


class NotebookId(ResourceId):
    pass


_WithNotebookSuffix = WithSuffix(".ipynb", case_sensitive=True)
NotebookSource = Union[
    Annotated[HttpUrl, _WithNotebookSuffix],
    Annotated[FilePath, _WithNotebookSuffix],
    Annotated[RelativeFilePath, _WithNotebookSuffix],
]


class NotebookDescr(GenericDescrBase):
    """Bioimage.io description of a Jupyter Notebook."""

    implemented_type: ClassVar[Literal["notebook"]] = "notebook"
    if TYPE_CHECKING:
        type: Literal["notebook"] = "notebook"
    else:
        type: Literal["notebook"]

    id: Optional[NotebookId] = None
    """bioimage.io-wide unique resource identifier
    assigned by bioimage.io; version **un**specific."""

    source: NotebookSource
    """The Jupyter notebook"""


class LinkedNotebook(Node):
    """Reference to a bioimage.io notebook."""

    id: NotebookId
    """A valid notebook `id` from the bioimage.io collection."""

    version_number: Optional[int] = None
    """version number (n-th published version, not the semantic version) of linked notebook"""



import warnings
from pprint import pformat
from types import TracebackType
from typing import Any, List, Type, Union

from pydantic import ValidationError

from .summary import format_loc

try:
    from IPython.core.getipython import get_ipython
    from IPython.core.interactiveshell import InteractiveShell

    class PrettyValidationError(ValueError):
        """Wrap a pydantic.ValidationError to custumize formatting."""

        def __init__(self, validation_error: ValidationError):
            super().__init__()
            self.error = validation_error

        def __str__(self):
            errors: List[str] = []
            for e in self.error.errors(include_url=False):
                ipt_lines = pformat(
                    e["input"], sort_dicts=False, depth=1, compact=True, width=30
                ).split("\n")
                if len(ipt_lines) > 2:
                    ipt_lines[1:-1] = ["..."]

                ipt = " ".join([il.strip() for il in ipt_lines])

                errors.append(
                    f"\n{format_loc(e['loc'], 'plain')}\n  {e['msg']} [input={ipt}]"
                )

            return (
                f"{self.error.error_count()} validation errors for"
                f" {self.error.title}:{''.join(errors)}"
            )

    def _custom_exception_handler(
        self: InteractiveShell,
        etype: Type[ValidationError],
        evalue: ValidationError,
        tb: TracebackType,
        tb_offset: Any = None,
    ):
        assert issubclass(etype, ValidationError), type(etype)
        assert isinstance(evalue, ValidationError), type(etype)

        stb: Union[Any, List[Union[str, Any]]]
        stb = self.InteractiveTB.structured_traceback(
            etype, PrettyValidationError(evalue), tb, tb_offset=tb_offset
        )

        if isinstance(stb, list):
            stb_clean = []
            for line in stb:  # pyright: ignore[reportUnknownVariableType]
                if (
                    isinstance(line, str)
                    and "pydantic" in line
                    and "__tracebackhide__" in line
                ):
                    # ignore pydantic internal frame in traceback
                    continue
                stb_clean.append(line)

            stb = stb_clean

        self._showtraceback(etype, PrettyValidationError(evalue), stb)  # type: ignore

    def _enable_pretty_validation_errors_in_ipynb():
        """A modestly hacky way to display prettified validaiton error messages and traceback
        in interactive Python notebooks"""
        ipy = get_ipython()
        if ipy is not None:
            ipy.set_custom_exc((ValidationError,), _custom_exception_handler)

except ImportError:
    _enabled = False
else:
    try:
        _enable_pretty_validation_errors_in_ipynb()
    except Exception as e:
        _enabled = False
        warnings.warn(
            "Failed to enable pretty validation errors in ipython: " + str(e),
            stacklevel=2,
        )
    else:
        _enabled = True

PRETTY_VALIDATION_ERRORS_IN_IPYNB_ENABLED = _enabled
"""Whether pretty validation errors in IPython notebooks were successfully enabled during import."""



from pathlib import Path


def test_save_markdown(tmp_path: Path):
    from bioimageio.spec.summary import ValidationDetail, ValidationSummary

    summary = ValidationSummary(
        type="model",
        format_version="0.5.5",
        name="test",
        source_name="source",
        status="passed",
        details=[ValidationDetail(name="test", status="passed")],
    )
    p = tmp_path / "out.md"
    summary.save_markdown(p)
    assert p.exists()


def test_summary_io(tmp_path: Path):
    from bioimageio.spec.summary import ValidationDetail, ValidationSummary

    summary = ValidationSummary(
        type="generic",
        format_version="0.3.0",
        name="test",
        source_name="source",
        status="failed",
        details=[ValidationDetail(name="test", status="passed")],
    )
    p = tmp_path / "summary.json"
    _ = summary.save(p)
    loaded_summary = summary.load_json(p)
    assert loaded_summary == summary



from __future__ import annotations

from contextvars import ContextVar, Token
from dataclasses import dataclass, field
from typing import Callable, Dict, List, Literal, Optional, Sequence, Union, cast

from .io import FileDescr
from .io_basics import FileName
from .utils import SLOTS


@dataclass(frozen=True, **SLOTS)
class PackagingContext:
    _context_tokens: "List[Token[Optional[PackagingContext]]]" = field(
        init=False,
        default_factory=cast(
            "Callable[[], List[Token[Optional[PackagingContext]]]]", list
        ),
    )

    bioimageio_yaml_file_name: FileName

    file_sources: Dict[FileName, FileDescr]
    """File sources to include in the packaged resource"""

    weights_priority_order: Optional[Sequence[str]] = None
    """set to select a single weights entry when packaging model resources"""

    def replace(
        self,
        *,
        bioimageio_yaml_file_name: Optional[FileName] = None,
        file_sources: Optional[Dict[FileName, FileDescr]] = None,
        weights_priority_order: Union[
            Optional[Sequence[str]], Literal["unchanged"]
        ] = "unchanged",
    ) -> "PackagingContext":
        """return a modiefied copy"""
        return PackagingContext(
            bioimageio_yaml_file_name=(
                self.bioimageio_yaml_file_name
                if bioimageio_yaml_file_name is None
                else bioimageio_yaml_file_name
            ),
            file_sources=(
                dict(self.file_sources) if file_sources is None else file_sources
            ),
            weights_priority_order=(
                self.weights_priority_order
                if weights_priority_order == "unchanged"
                else weights_priority_order
            ),
        )

    def __enter__(self):
        self._context_tokens.append(packaging_context_var.set(self))
        return self

    def __exit__(self, type, value, traceback):  # type: ignore
        packaging_context_var.reset(self._context_tokens.pop(-1))


packaging_context_var: ContextVar[Optional[PackagingContext]] = ContextVar(
    "packaging_context_var", default=None
)



"""Representation of conda environment.yaml files for bioimageio specifications."""

import warnings
from typing import Any, Callable, List, Optional, Union, cast

from pydantic import BaseModel, Field, field_validator, model_validator


class PipDeps(BaseModel):
    """Pip dependencies to include in conda dependecies"""

    pip: List[str] = Field(default_factory=list)

    @field_validator("pip", mode="after")
    @classmethod
    def _remove_empty_and_sort(cls, value: List[str]) -> List[str]:
        return sorted((vs for v in value if (vs := v.strip())))

    def __lt__(self, other: Any):
        if isinstance(other, PipDeps):
            return len(self.pip) < len(other.pip)
        else:
            return False

    def __gt__(self, other: Any):
        if isinstance(other, PipDeps):
            return len(self.pip) > len(other.pip)
        else:
            return False


class CondaEnv(BaseModel):
    """Represenation of the content of a conda environment.yaml file"""

    name: Optional[str] = None
    channels: List[str] = Field(default_factory=list)
    dependencies: List[Union[str, PipDeps]] = Field(
        default_factory=cast(Callable[[], List[Union[str, PipDeps]]], list)
    )

    @field_validator("name", mode="after")
    def _ensure_valid_conda_env_name(cls, value: Optional[str]) -> Optional[str]:
        if value is None:
            return None

        for illegal in ("/", " ", ":", "#"):
            value = value.replace(illegal, "")

        return value or "empty"

    @property
    def wo_name(self):
        return self.model_construct(**{k: v for k, v in self if k != "name"})

    def _get_version_pin(self, package: str):
        """Helper to return any version pin for **package**

        TODO: improve: interprete version pin and return structured information.
        """
        for d in self.dependencies:
            if isinstance(d, PipDeps):
                for p in d.pip:
                    if p.startswith(package):
                        return p[len(package) :]
            elif d.startswith(package):
                return d[len(package) :]
            elif "::" in d and (d_wo_channel := d.split("::", 1)[-1]).startswith(
                package
            ):
                return d_wo_channel[len(package) :]

    def get_pip_deps(self) -> List[str]:
        """Get the pip dependencies of this conda env."""
        for dep in self.dependencies:
            if isinstance(dep, PipDeps):
                return dep.pip

        return []


class BioimageioCondaEnv(CondaEnv):
    """A special `CondaEnv` that
    - automatically adds bioimageio specific dependencies
    - sorts dependencies
    """

    @model_validator(mode="after")
    def _normalize_bioimageio_conda_env(self):
        """update a conda env such that we have bioimageio.core and sorted dependencies"""
        for req_channel in ("conda-forge", "nodefaults"):
            if req_channel not in self.channels:
                self.channels.append(req_channel)

        if "defaults" in self.channels:
            warnings.warn("removing 'defaults' from conda-channels")
            self.channels.remove("defaults")

        if "pip" not in self.dependencies:
            self.dependencies.append("pip")

        for dep in self.dependencies:
            if isinstance(dep, PipDeps):
                pip_section = dep
                pip_section.pip.sort()
                break
        else:
            pip_section = None

        if (
            pip_section is None
            or not any(pd.startswith("bioimageio.core") for pd in pip_section.pip)
        ) and not any(
            d.startswith("bioimageio.core")
            or d.startswith("conda-forge::bioimageio.core")
            for d in self.dependencies
            if not isinstance(d, PipDeps)
        ):
            self.dependencies.append("conda-forge::bioimageio.core")

        self.dependencies.sort()
        return self



# autogen: start
"""model resource description format

Implementations of all released minor versions are available in submodules.
"""

from typing import Union

from pydantic import Discriminator, Field
from typing_extensions import Annotated

from . import v0_4, v0_5

ModelDescr = v0_5.ModelDescr
ModelDescr_v0_4 = v0_4.ModelDescr
ModelDescr_v0_5 = v0_5.ModelDescr

AnyModelDescr = Annotated[
    Union[
        Annotated[ModelDescr_v0_4, Field(title="model 0.4")],
        Annotated[ModelDescr_v0_5, Field(title="model 0.5")],
    ],
    Discriminator("format_version"),
    Field(title="model"),
]
"""Union of any released model desription"""
# autogen: stop



from types import ModuleType
from typing import Any, Dict, Mapping

import pytest

from bioimageio.spec import application, dataset, generic, model, notebook

IGNORE_MEMBERS = {
    "AfterValidator",
    "ALERT",
    "annotated_types",
    "Annotated",
    "annotations",
    "Any",
    "as_warning",
    "assert_never",
    "BioimageioYamlContent",
    "Callable",
    "cast",
    "ClassVar",
    "collections",
    "convert_from_older_format",
    "Converter",
    "CoverImageSource",
    "DeprecatedLicenseId",
    "Dict",
    "EmailStr",
    "FAIR",
    "field_validator",
    "Field",
    "FilePath",
    "FileSource",
    "Ge",
    "GenericDescr",
    "GenericDescrBase",
    "GenericModelDescrBase",
    "get_args",
    "get_validation_context",
    "httpx",
    "include_in_package_serializer",
    "include_in_package",
    "INFO",
    "is_dict",
    "is_sequence",
    "is_yaml_value",
    "issue_warning",
    "KNOWN_SPECIFIC_RESOURCE_TYPES",
    "Len",
    "LicenseId",
    "LICENSES",
    "LinkedResourceBase",
    "List",
    "LowerCase",
    "Mapping",
    "MarkdownSource",
    "MaxLen",
    "MinLen",
    "model_validator",
    "Node",
    "NotEmpty",
    "Optional",
    "partial",
    "PermissiveFileSource",
    "Predicate",
    "pydantic",
    "ResourceDescrBase",
    "ResourceDescrType",
    "RestrictCharacters",
    "RootModel",
    "S",
    "Self",
    "Sequence",
    "settings",
    "string",
    "TAG_CATEGORIES",
    "Type",
    "TypeVar",
    "Union",
    "V_suffix",
    "v0_2",
    "v0_3",
    "v0_4",
    "v0_5",
    "validate_github_user",
    "validate_suffix",
    "ValidatedString",
    "ValidationInfo",
    "warn",
    "WithSuffix",
    "wo_special_file_name",
    "YamlValue",
}


def get_members(m: ModuleType) -> Mapping[str, Any]:
    return {
        k: getattr(m, k)
        for k in dir(m)
        if not k.startswith("_")
        and k not in IGNORE_MEMBERS
        and not k.startswith("FileSource_")  # Annotated[FileSource, ...]
        and not k.startswith("FileDescr_")  # Annotated[FileDescr, ...]
    }


GENERIC_v0_2_MEMBERS = get_members(generic.v0_2)
GENERIC_v0_3_MEMBERS = get_members(generic.v0_3)


@pytest.mark.parametrize(
    "generic_members,specific",
    [
        (GENERIC_v0_2_MEMBERS, application.v0_2),
        (GENERIC_v0_2_MEMBERS, dataset.v0_2),
        (GENERIC_v0_2_MEMBERS, model.v0_4),
        (GENERIC_v0_2_MEMBERS, notebook.v0_2),
        (GENERIC_v0_3_MEMBERS, application.v0_3),
        (GENERIC_v0_3_MEMBERS, dataset.v0_3),
        (GENERIC_v0_3_MEMBERS, model.v0_5),
        (GENERIC_v0_3_MEMBERS, notebook.v0_3),
    ],
)
def test_specific_module_has_all_generic_symbols(
    generic_members: Dict[str, Any], specific: ModuleType
):
    members = get_members(specific)
    missing = {k for k in generic_members if k not in members}
    assert not missing



import collections.abc
from pathlib import Path
from typing import Dict, Literal, Optional, TextIO, Union, cast, overload
from zipfile import ZipFile

from loguru import logger
from pydantic import FilePath, NewPath

from ._description import (
    DISCOVER,
    LATEST,
    InvalidDescr,
    LatestResourceDescr,
    ResourceDescr,
    build_description,
    dump_description,
    ensure_description_is_dataset,
    ensure_description_is_model,
)
from ._internal.common_nodes import ResourceDescrBase
from ._internal.io import BioimageioYamlContent, YamlValue
from ._internal.io_basics import Sha256
from ._internal.io_utils import open_bioimageio_yaml, write_yaml
from ._internal.types import FormatVersionPlaceholder, PermissiveFileSource
from ._internal.validation_context import get_validation_context
from .dataset import AnyDatasetDescr, DatasetDescr
from .model import AnyModelDescr, ModelDescr
from .summary import ValidationSummary


@overload
def load_description(
    source: Union[PermissiveFileSource, ZipFile],
    /,
    *,
    format_version: Literal["latest"],
    perform_io_checks: Optional[bool] = None,
    known_files: Optional[Dict[str, Optional[Sha256]]] = None,
    sha256: Optional[Sha256] = None,
) -> Union[LatestResourceDescr, InvalidDescr]: ...


@overload
def load_description(
    source: Union[PermissiveFileSource, ZipFile],
    /,
    *,
    format_version: Union[FormatVersionPlaceholder, str] = DISCOVER,
    perform_io_checks: Optional[bool] = None,
    known_files: Optional[Dict[str, Optional[Sha256]]] = None,
    sha256: Optional[Sha256] = None,
) -> Union[ResourceDescr, InvalidDescr]: ...


def load_description(
    source: Union[PermissiveFileSource, ZipFile],
    /,
    *,
    format_version: Union[FormatVersionPlaceholder, str] = DISCOVER,
    perform_io_checks: Optional[bool] = None,
    known_files: Optional[Dict[str, Optional[Sha256]]] = None,
    sha256: Optional[Sha256] = None,
) -> Union[ResourceDescr, InvalidDescr]:
    """load a bioimage.io resource description

    Args:
        source:
            Path or URL to an rdf.yaml or a bioimage.io package
            (zip-file with rdf.yaml in it).
        format_version:
            (optional) Use this argument to load the resource and
            convert its metadata to a higher format_version.
            Note:
            - Use "latest" to convert to the latest available format version.
            - Use "discover" to use the format version specified in the RDF.
            - Only considers major.minor format version, ignores patch version.
            - Conversion to lower format versions is not supported.
        perform_io_checks:
            Wether or not to perform validation that requires file io,
            e.g. downloading a remote files. The existence of local
            absolute file paths is still being checked.
        known_files:
            Allows to bypass download and hashing of referenced files
            (even if perform_io_checks is True).
            Checked files will be added to this dictionary
            with their SHA-256 value.
        sha256:
            Optional SHA-256 value of **source**

    Returns:
        An object holding all metadata of the bioimage.io resource

    """
    if isinstance(source, ResourceDescrBase):
        name = getattr(source, "name", f"{str(source)[:10]}...")
        logger.warning("returning already loaded description '{}' as is", name)
        return source  # pyright: ignore[reportReturnType]

    opened = open_bioimageio_yaml(source, sha256=sha256)

    context = get_validation_context().replace(
        root=opened.original_root,
        file_name=opened.original_file_name,
        original_source_name=opened.original_source_name,
        perform_io_checks=perform_io_checks,
        known_files=known_files,
    )

    return build_description(
        opened.content,
        context=context,
        format_version=format_version,
    )


@overload
def load_model_description(
    source: Union[PermissiveFileSource, ZipFile],
    /,
    *,
    format_version: Literal["latest"],
    perform_io_checks: Optional[bool] = None,
    known_files: Optional[Dict[str, Optional[Sha256]]] = None,
    sha256: Optional[Sha256] = None,
) -> ModelDescr: ...


@overload
def load_model_description(
    source: Union[PermissiveFileSource, ZipFile],
    /,
    *,
    format_version: Union[FormatVersionPlaceholder, str] = DISCOVER,
    perform_io_checks: Optional[bool] = None,
    known_files: Optional[Dict[str, Optional[Sha256]]] = None,
    sha256: Optional[Sha256] = None,
) -> AnyModelDescr: ...


def load_model_description(
    source: Union[PermissiveFileSource, ZipFile],
    /,
    *,
    format_version: Union[FormatVersionPlaceholder, str] = DISCOVER,
    perform_io_checks: Optional[bool] = None,
    known_files: Optional[Dict[str, Optional[Sha256]]] = None,
    sha256: Optional[Sha256] = None,
) -> AnyModelDescr:
    """same as `load_description`, but addtionally ensures that the loaded
    description is valid and of type 'model'.

    Raises:
        ValueError: for invalid or non-model resources
    """
    rd = load_description(
        source,
        format_version=format_version,
        perform_io_checks=perform_io_checks,
        known_files=known_files,
        sha256=sha256,
    )
    return ensure_description_is_model(rd)


@overload
def load_dataset_description(
    source: Union[PermissiveFileSource, ZipFile],
    /,
    *,
    format_version: Literal["latest"],
    perform_io_checks: Optional[bool] = None,
    known_files: Optional[Dict[str, Optional[Sha256]]] = None,
    sha256: Optional[Sha256] = None,
) -> DatasetDescr: ...


@overload
def load_dataset_description(
    source: Union[PermissiveFileSource, ZipFile],
    /,
    *,
    format_version: Union[FormatVersionPlaceholder, str] = DISCOVER,
    perform_io_checks: Optional[bool] = None,
    known_files: Optional[Dict[str, Optional[Sha256]]] = None,
    sha256: Optional[Sha256] = None,
) -> AnyDatasetDescr: ...


def load_dataset_description(
    source: Union[PermissiveFileSource, ZipFile],
    /,
    *,
    format_version: Union[FormatVersionPlaceholder, str] = DISCOVER,
    perform_io_checks: Optional[bool] = None,
    known_files: Optional[Dict[str, Optional[Sha256]]] = None,
    sha256: Optional[Sha256] = None,
) -> AnyDatasetDescr:
    """same as `load_description`, but addtionally ensures that the loaded
    description is valid and of type 'dataset'.
    """
    rd = load_description(
        source,
        format_version=format_version,
        perform_io_checks=perform_io_checks,
        known_files=known_files,
        sha256=sha256,
    )
    return ensure_description_is_dataset(rd)


def save_bioimageio_yaml_only(
    rd: Union[ResourceDescr, BioimageioYamlContent, InvalidDescr],
    /,
    file: Union[NewPath, FilePath, TextIO],
    *,
    exclude_unset: bool = True,
    exclude_defaults: bool = False,
):
    """write the metadata of a resource description (`rd`) to `file`
    without writing any of the referenced files in it.

    Args:
        rd: bioimageio resource description
        file: file or stream to save to
        exclude_unset: Exclude fields that have not explicitly be set.
        exclude_defaults: Exclude fields that have the default value (even if set explicitly).

    Note: To save a resource description with its associated files as a package,
    use `save_bioimageio_package` or `save_bioimageio_package_as_folder`.
    """
    if isinstance(rd, ResourceDescrBase):
        content = dump_description(
            rd, exclude_unset=exclude_unset, exclude_defaults=exclude_defaults
        )
    else:
        content = rd

    write_yaml(cast(YamlValue, content), file)


def load_description_and_validate_format_only(
    source: Union[PermissiveFileSource, ZipFile],
    /,
    *,
    format_version: Union[FormatVersionPlaceholder, str] = DISCOVER,
    perform_io_checks: Optional[bool] = None,
    known_files: Optional[Dict[str, Optional[Sha256]]] = None,
    sha256: Optional[Sha256] = None,
) -> ValidationSummary:
    """same as `load_description`, but only return the validation summary.

    Returns:
        Validation summary of the bioimage.io resource found at `source`.

    """
    rd = load_description(
        source,
        format_version=format_version,
        perform_io_checks=perform_io_checks,
        known_files=known_files,
        sha256=sha256,
    )
    assert rd.validation_summary is not None
    return rd.validation_summary


def update_format(
    source: Union[
        ResourceDescr,
        PermissiveFileSource,
        ZipFile,
        BioimageioYamlContent,
        InvalidDescr,
    ],
    /,
    *,
    output: Union[Path, TextIO, None] = None,
    exclude_defaults: bool = True,
    perform_io_checks: Optional[bool] = None,
) -> Union[LatestResourceDescr, InvalidDescr]:
    """Update a resource description.

    Notes:
    - Invalid **source** descriptions may fail to update.
    - The updated description might be invalid (even if the **source** was valid).
    """

    if isinstance(source, ResourceDescrBase):
        root = source.root
        source = dump_description(source)
    else:
        root = None

    if isinstance(source, collections.abc.Mapping):
        descr = build_description(
            source,
            context=get_validation_context().replace(
                root=root, perform_io_checks=perform_io_checks
            ),
            format_version=LATEST,
        )

    else:
        descr = load_description(
            source,
            perform_io_checks=perform_io_checks,
            format_version=LATEST,
        )

    if output is not None:
        save_bioimageio_yaml_only(descr, file=output, exclude_defaults=exclude_defaults)

    return descr


def update_hashes(
    source: Union[PermissiveFileSource, ZipFile, ResourceDescr, BioimageioYamlContent],
    /,
) -> Union[ResourceDescr, InvalidDescr]:
    """Update hash values of the files referenced in **source**."""
    if isinstance(source, ResourceDescrBase):
        root = source.root
        source = dump_description(source)
    else:
        root = None

    context = get_validation_context().replace(
        update_hashes=True, root=root, perform_io_checks=True
    )
    with context:
        if isinstance(source, collections.abc.Mapping):
            return build_description(source)
        else:
            return load_description(source, perform_io_checks=True)



from __future__ import annotations

from abc import ABC
from io import BytesIO
from pathlib import Path
from types import MappingProxyType
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    Protocol,
    Tuple,
    Union,
)
from zipfile import ZipFile

import pydantic
from pydantic import DirectoryPath, PrivateAttr, model_validator
from pydantic_core import PydanticUndefined
from typing_extensions import Self

from ..summary import (
    WARNING_LEVEL_TO_NAME,
    ErrorEntry,
    ValidationDetail,
    ValidationSummary,
    WarningEntry,
)
from .field_warning import issue_warning
from .io import (
    BioimageioYamlContent,
    BioimageioYamlContentView,
    FileDescr,
    deepcopy_yaml_value,
    extract_file_descrs,
    populate_cache,
)
from .io_basics import BIOIMAGEIO_YAML, FileName
from .io_utils import write_content_to_zip
from .node import Node
from .packaging_context import PackagingContext
from .root_url import RootHttpUrl
from .type_guards import is_dict
from .utils import get_format_version_tuple
from .validation_context import ValidationContext, get_validation_context
from .warning_levels import ALERT, ERROR, INFO


class NodeWithExplicitlySetFields(Node):
    _fields_to_set_explicitly: ClassVar[Mapping[str, Any]]

    @classmethod
    def __pydantic_init_subclass__(cls, **kwargs: Any) -> None:
        explict_fields: Dict[str, Any] = {}
        for attr in dir(cls):
            if attr.startswith("implemented_"):
                field_name = attr.replace("implemented_", "")
                if field_name not in cls.model_fields:
                    continue

                assert (
                    cls.model_fields[field_name].get_default() is PydanticUndefined
                ), field_name
                default = getattr(cls, attr)
                explict_fields[field_name] = default

        cls._fields_to_set_explicitly = MappingProxyType(explict_fields)
        return super().__pydantic_init_subclass__(**kwargs)

    @model_validator(mode="before")
    @classmethod
    def _set_fields_explicitly(
        cls, data: Union[Any, Dict[str, Any]]
    ) -> Union[Any, Dict[str, Any]]:
        if isinstance(data, dict):
            for name, default in cls._fields_to_set_explicitly.items():
                if name not in data:
                    data[name] = default

        return data  # pyright: ignore[reportUnknownVariableType]


if TYPE_CHECKING:

    class _ResourceDescrBaseAbstractFieldsProtocol(Protocol):
        """workaround to add "abstract" fields to ResourceDescrBase"""

        # TODO: implement as proper abstract fields of ResourceDescrBase

        type: Any  # should be LiteralString
        format_version: Any  # should be LiteralString
        implemented_type: ClassVar[Any]
        implemented_format_version: ClassVar[Any]

else:

    class _ResourceDescrBaseAbstractFieldsProtocol:
        pass


class ResourceDescrBase(
    NodeWithExplicitlySetFields, ABC, _ResourceDescrBaseAbstractFieldsProtocol
):
    """base class for all resource descriptions"""

    _validation_summary: Optional[ValidationSummary] = None

    implemented_format_version_tuple: ClassVar[Tuple[int, int, int]]

    # @field_validator("format_version", mode="before", check_fields=False)
    # field_validator on "format_version" is not possible, because we want to use
    #   "format_version" in a descriminated Union higher up
    # (PydanticUserError: Cannot use a mode='before' validator in the discriminator
    #   field 'format_version' of Model 'CollectionDescr')
    @model_validator(mode="before")
    @classmethod
    def _ignore_future_patch(cls, data: Any, /) -> Any:
        if (
            cls.implemented_format_version == "unknown"
            or not is_dict(data)
            or "format_version" not in data
        ):
            return data

        value = data["format_version"]
        fv = get_format_version_tuple(value)
        if fv is None:
            return data
        if (
            fv[0] == cls.implemented_format_version_tuple[0]
            and fv[1:] > cls.implemented_format_version_tuple[1:]
        ):
            issue_warning(
                "future format_version '{value}' treated as '{implemented}'",
                value=value,
                msg_context=dict(implemented=cls.implemented_format_version),
                severity=ALERT,
            )
            data["format_version"] = cls.implemented_format_version

        return data

    @model_validator(mode="after")
    def _set_init_validation_summary(self) -> Self:
        context = get_validation_context()

        self._validation_summary = ValidationSummary(
            name="bioimageio format validation",
            source_name=context.source_name,
            id=getattr(self, "id", None),
            type=self.type,
            format_version=self.format_version,
            status="failed" if isinstance(self, InvalidDescr) else "valid-format",
            metadata_completeness=self._get_metadata_completeness(),
            details=(
                []
                if isinstance(self, InvalidDescr)
                else [
                    ValidationDetail(
                        name=f"Successfully created `{self.__class__.__name__}` instance.",
                        status="passed",
                        context=context.summary,
                    )
                ]
            ),
        )
        return self

    @property
    def validation_summary(self) -> ValidationSummary:
        assert self._validation_summary is not None, "access only after initialization"
        return self._validation_summary

    _root: Union[RootHttpUrl, DirectoryPath, ZipFile] = PrivateAttr(
        default_factory=lambda: get_validation_context().root
    )

    _file_name: Optional[FileName] = PrivateAttr(
        default_factory=lambda: get_validation_context().file_name
    )

    @property
    def root(self) -> Union[RootHttpUrl, DirectoryPath, ZipFile]:
        """The URL/Path prefix to resolve any relative paths with."""
        return self._root

    @property
    def file_name(self) -> Optional[FileName]:
        """File name of the bioimageio.yaml file the description was loaded from."""
        return self._file_name

    @classmethod
    def __pydantic_init_subclass__(cls, **kwargs: Any):
        super().__pydantic_init_subclass__(**kwargs)
        # set classvar implemented_format_version_tuple
        if "format_version" in cls.model_fields:
            if "." not in cls.implemented_format_version:
                cls.implemented_format_version_tuple = (0, 0, 0)
            else:
                fv_tuple = get_format_version_tuple(cls.implemented_format_version)
                assert fv_tuple is not None, (
                    f"failed to cast '{cls.implemented_format_version}' to tuple"
                )
                cls.implemented_format_version_tuple = fv_tuple

    @classmethod
    def load(
        cls,
        data: BioimageioYamlContentView,
        context: Optional[ValidationContext] = None,
    ) -> Union[Self, InvalidDescr]:
        """factory method to create a resource description object"""
        context = context or get_validation_context()
        if context.perform_io_checks:
            file_descrs = extract_file_descrs({k: v for k, v in data.items()})
            populate_cache(file_descrs)  # TODO: add progress bar

        with context.replace(log_warnings=context.warning_level <= INFO):
            rd, errors, val_warnings = cls._load_impl(deepcopy_yaml_value(data))

        if context.warning_level > INFO:
            all_warnings_context = context.replace(
                warning_level=INFO, log_warnings=False, raise_errors=False
            )
            # raise all validation warnings by reloading
            with all_warnings_context:
                _, _, val_warnings = cls._load_impl(deepcopy_yaml_value(data))

        format_status = "failed" if errors else "passed"
        rd.validation_summary.add_detail(
            ValidationDetail(
                errors=errors,
                name=(
                    "bioimageio.spec format validation"
                    f" {rd.type} {cls.implemented_format_version}"
                ),
                status=format_status,
                warnings=val_warnings,
            ),
            update_status=False,  # avoid updating status from 'valid-format' to 'passed', but ...
        )
        if format_status == "failed":
            # ... update status in case of failure
            rd.validation_summary.status = "failed"

        return rd

    def _get_metadata_completeness(self) -> float:
        if isinstance(self, InvalidDescr):
            return 0.0

        given = self.model_dump(mode="json", exclude_unset=True, exclude_defaults=False)
        full = self.model_dump(mode="json", exclude_unset=False, exclude_defaults=False)

        def extract_flat_keys(d: Dict[Any, Any], key: str = "") -> Iterable[str]:
            for k, v in d.items():
                if is_dict(v):
                    yield from extract_flat_keys(v, key=f"{key}.{k}" if key else k)

                yield f"{key}.{k}" if key else k

        given_keys = set(extract_flat_keys(given))
        full_keys = set(extract_flat_keys(full))
        assert len(full_keys) >= len(given_keys)
        return len(given_keys) / len(full_keys) if full_keys else 0.0

    @classmethod
    def _load_impl(
        cls, data: BioimageioYamlContent
    ) -> Tuple[Union[Self, InvalidDescr], List[ErrorEntry], List[WarningEntry]]:
        rd: Union[Self, InvalidDescr, None] = None
        val_errors: List[ErrorEntry] = []
        val_warnings: List[WarningEntry] = []

        context = get_validation_context()
        try:
            rd = cls.model_validate(data)
        except pydantic.ValidationError as e:
            for ee in e.errors(include_url=False):
                if (severity := ee.get("ctx", {}).get("severity", ERROR)) < ERROR:
                    val_warnings.append(
                        WarningEntry(
                            loc=ee["loc"],
                            msg=ee["msg"],
                            type=ee["type"],
                            severity=severity,
                        )
                    )
                elif context.raise_errors:
                    raise e
                else:
                    val_errors.append(
                        ErrorEntry(loc=ee["loc"], msg=ee["msg"], type=ee["type"])
                    )

            if len(val_errors) == 0:  # FIXME is this reduntant?
                val_errors.append(
                    ErrorEntry(
                        loc=(),
                        msg=(
                            f"Encountered {len(val_warnings)} more severe than warning"
                            " level "
                            f"'{WARNING_LEVEL_TO_NAME[context.warning_level]}'"
                        ),
                        type="severe_warnings",
                    )
                )
        except Exception as e:
            if context.raise_errors:
                raise e

            try:
                msg = str(e)
            except Exception:
                msg = e.__class__.__name__ + " encountered"

            val_errors.append(
                ErrorEntry(
                    loc=(),
                    msg=msg,
                    type=type(e).__name__,
                    with_traceback=True,
                )
            )

        if rd is None:
            try:
                rd = InvalidDescr.model_validate(data)
            except Exception as e:
                if context.raise_errors:
                    raise e
                resource_type = cls.model_fields["type"].default
                format_version = cls.implemented_format_version
                rd = InvalidDescr(type=resource_type, format_version=format_version)
                if context.raise_errors:
                    raise ValueError(rd)

        return rd, val_errors, val_warnings

    def package(
        self, dest: Optional[Union[ZipFile, IO[bytes], Path, str]] = None, /
    ) -> ZipFile:
        """package the described resource as a zip archive

        Args:
            dest: (path/bytes stream of) destination zipfile
        """
        if dest is None:
            dest = BytesIO()

        if isinstance(dest, ZipFile):
            zip = dest
            if "r" in zip.mode:
                raise ValueError(
                    f"zip file {dest} opened in '{zip.mode}' mode,"
                    + " but write access is needed for packaging."
                )
        else:
            zip = ZipFile(dest, mode="w")

        if zip.filename is None:
            zip.filename = (
                str(getattr(self, "id", getattr(self, "name", "bioimageio"))) + ".zip"
            )

        content = self.get_package_content()
        write_content_to_zip(content, zip)
        return zip

    def get_package_content(
        self,
    ) -> Dict[FileName, Union[FileDescr, BioimageioYamlContent]]:
        """Returns package content without creating the package."""
        content: Dict[FileName, FileDescr] = {}
        with PackagingContext(
            bioimageio_yaml_file_name=BIOIMAGEIO_YAML,
            file_sources=content,
        ):
            rdf_content: BioimageioYamlContent = self.model_dump(
                mode="json", exclude_unset=True
            )

        _ = rdf_content.pop("rdf_source", None)

        return {**content, BIOIMAGEIO_YAML: rdf_content}


class InvalidDescr(
    ResourceDescrBase,
    extra="allow",
    title="An invalid resource description",
):
    """A representation of an invalid resource description"""

    implemented_type: ClassVar[Literal["unknown"]] = "unknown"
    if TYPE_CHECKING:  # see NodeWithExplicitlySetFields
        type: Any = "unknown"
    else:
        type: Any

    implemented_format_version: ClassVar[Literal["unknown"]] = "unknown"
    if TYPE_CHECKING:  # see NodeWithExplicitlySetFields
        format_version: Any = "unknown"
    else:
        format_version: Any


class KwargsNode(Node):
    def get(self, item: str, default: Any = None) -> Any:
        return self[item] if item in self else default

    def __getitem__(self, item: str) -> Any:
        if item in self.__class__.model_fields:
            return getattr(self, item)
        else:
            raise KeyError(item)

    def __contains__(self, item: str) -> bool:
        return item in self.__class__.model_fields



from pathlib import Path
from typing import Any, Dict, List, Mapping, Optional, Union

import pytest

from bioimageio.spec._internal.validation_context import ValidationContext
from bioimageio.spec.model.v0_5 import (
    OnnxWeightsDescr,
    PytorchStateDictWeightsDescr,
    TorchscriptWeightsDescr,
)


@pytest.mark.parametrize(
    "descr_class,w",
    [
        (
            PytorchStateDictWeightsDescr,
            dict(
                authors=[
                    dict(
                        name="Constantin Pape;@bioimage-io",
                        affiliation="EMBL Heidelberg",
                        orcid="0000-0001-6562-7187",
                    )
                ],
                sha256="e4d3885bccbe41cbf6c1d825f3cd2b707c7021ead5593156007e407a16b27cf2",
                source="http://files.ilastik.org/bioimageio_test/unet2d_nuclei_broad_weights.torch",
                architecture=dict(
                    callable="UNet2d",
                    source="unet2d.py",
                    sha256="7cdd8332dc3e3735e71c328f81b63a9ac86c028f80522312484ca9a4027d4ce1",
                    kwargs=dict(input_channels=1, output_channels=1),
                ),
                dependencies=dict(
                    source="environment.yaml",
                    sha256="129d589d2ec801398719b1a6d1bf20ea36b3632f14ccb56a24700df7d719fd10",
                ),
                pytorch_version="1.5.1",
            ),
        ),
        (
            OnnxWeightsDescr,
            dict(
                sha256="f1f086d5e340f9d4d7001a1b62a2b835f9b87a2fb5452c4fe7d8cc821bdf539c",
                source="weights.onnx",
                opset_version=12,
                parent="pytorch_state_dict",
            ),
        ),
        (
            TorchscriptWeightsDescr,
            dict(
                sha256="62fa1c39923bee7d58a192277e0dd58f2da9ee810662addadd0f44a3784d9210",
                source="weights.pt",
                parent="pytorch_state_dict",
                pytorch_version="1.5.1",
            ),
        ),
    ],
)
def test_get_conda_env(
    descr_class: Union[
        PytorchStateDictWeightsDescr, OnnxWeightsDescr, TorchscriptWeightsDescr
    ],
    w: Mapping[str, Any],
    unet2d_path: Path,
):
    from bioimageio.spec import get_conda_env

    with ValidationContext(perform_io_checks=False, root=unet2d_path.parent):
        w_descr = descr_class.model_validate(w)

    conda_env = get_conda_env(entry=w_descr)

    assert conda_env.channels
    assert conda_env.dependencies


def test_get_default_pytorch_env():
    from bioimageio.spec._get_conda_env import (
        _get_default_pytorch_env,  # pyright: ignore[reportPrivateUsage]
    )
    from bioimageio.spec._internal.version_type import Version

    versions: Dict[str, List[Optional[str]]] = {
        "pytorch": [
            "1.5.1",
            "1.6.0",
            "1.7.0",
            "1.7.1",
            "1.8.0",
            "1.8.1",
            "1.9.0",
            "1.9.1",
            "1.10.0",
            "1.10.1",
            "1.11.0",
            "1.12.0",
            "1.12.1",
            "1.13.0",
            "1.13.1",
            "2.0.0",
            "2.0.1",
            "2.1.0",
            "2.1.1",
            "2.1.2",
            "2.2.0",
            "2.2.1",
            "2.2.2",
            "2.3.0",
            "2.3.1",
            "2.4.0",
            "2.4.1",
            "2.5.0",
        ]
    }
    envs = [
        _get_default_pytorch_env(pytorch_version=Version.model_validate(v))
        for v in versions["pytorch"]
    ]
    for p in ["torchvision", "torchaudio"]:
        versions[p] = [
            env._get_version_pin(p)  # pyright: ignore[reportPrivateUsage]
            for env in envs
        ]

    def assert_lt(p: str, i: int):
        vs = versions[p]
        a, b = vs[i], vs[i + 1]
        assert a is not None, (vs[i], vs[i + 1])
        assert b is not None, (vs[i], vs[i + 1])
        av = Version(a.strip("="))
        bv = Version(b.strip("="))
        assert av < bv, (vs[i], vs[i + 1])

    for i in range(len(versions["pytorch"]) - 1):
        assert_lt("pytorch", i)
        assert_lt("torchvision", i)
        if i > 1:
            assert_lt("torchaudio", i)



from typing import TYPE_CHECKING, ClassVar, Literal, Optional

from pydantic import Field
from typing_extensions import Annotated

from .._internal.io import FileDescr as FileDescr
from .._internal.io_basics import Sha256 as Sha256
from .._internal.types import FAIR, FileSource_
from .._internal.url import HttpUrl as HttpUrl
from ..generic.v0_3 import VALID_COVER_IMAGE_EXTENSIONS as VALID_COVER_IMAGE_EXTENSIONS
from ..generic.v0_3 import Author as Author
from ..generic.v0_3 import BadgeDescr as BadgeDescr
from ..generic.v0_3 import BioimageioConfig as BioimageioConfig
from ..generic.v0_3 import CiteEntry as CiteEntry
from ..generic.v0_3 import Config as Config
from ..generic.v0_3 import DeprecatedLicenseId as DeprecatedLicenseId
from ..generic.v0_3 import Doi as Doi
from ..generic.v0_3 import GenericDescrBase, LinkedResourceBase, ResourceId
from ..generic.v0_3 import LicenseId as LicenseId
from ..generic.v0_3 import LinkedResource as LinkedResource
from ..generic.v0_3 import Maintainer as Maintainer
from ..generic.v0_3 import OrcidId as OrcidId
from ..generic.v0_3 import RelativeFilePath as RelativeFilePath
from ..generic.v0_3 import Uploader as Uploader
from ..generic.v0_3 import Version as Version


class ApplicationId(ResourceId):
    pass


class ApplicationDescr(GenericDescrBase):
    """Bioimage.io description of an application."""

    implemented_type: ClassVar[Literal["application"]] = "application"
    if TYPE_CHECKING:
        type: Literal["application"] = "application"
    else:
        type: Literal["application"]

    id: Optional[ApplicationId] = None
    """bioimage.io-wide unique resource identifier
    assigned by bioimage.io; version **un**specific."""

    parent: Optional[ApplicationId] = None
    """The description from which this one is derived"""

    source: Annotated[
        FAIR[Optional[FileSource_]],
        Field(description="URL or path to the source of the application"),
    ] = None
    """The primary source of the application"""


class LinkedApplication(LinkedResourceBase):
    """Reference to a bioimage.io application."""

    id: ApplicationId
    """A valid application `id` from the bioimage.io collection."""



from __future__ import annotations

import collections.abc
import hashlib
import sys
import warnings
import zipfile
from abc import abstractmethod
from contextlib import nullcontext
from dataclasses import dataclass, field
from datetime import date as _date
from datetime import datetime as _datetime
from functools import partial
from io import TextIOWrapper
from pathlib import Path, PurePath, PurePosixPath
from tempfile import mkdtemp
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypedDict,
    TypeVar,
    Union,
    overload,
)
from urllib.parse import urlparse, urlsplit, urlunsplit
from zipfile import ZipFile

import httpx
import pydantic
from genericache import NoopCache
from genericache.digest import ContentDigest, UrlDigest
from pydantic import (
    AnyUrl,
    DirectoryPath,
    Field,
    GetCoreSchemaHandler,
    PrivateAttr,
    RootModel,
    TypeAdapter,
    model_serializer,
    model_validator,
)
from pydantic_core import core_schema
from tqdm import tqdm
from typing_extensions import (
    Annotated,
    LiteralString,
    NotRequired,
    Self,
    TypeGuard,
    Unpack,
    assert_never,
)
from typing_extensions import TypeAliasType as _TypeAliasType

from ._settings import settings
from .io_basics import (
    ALL_BIOIMAGEIO_YAML_NAMES,
    ALTERNATIVE_BIOIMAGEIO_YAML_NAMES,
    BIOIMAGEIO_YAML,
    AbsoluteDirectory,
    AbsoluteFilePath,
    BytesReader,
    FileName,
    FilePath,
    Sha256,
    ZipPath,
    get_sha256,
)
from .node import Node
from .progress import Progressbar
from .root_url import RootHttpUrl
from .type_guards import is_dict, is_list, is_mapping, is_sequence
from .url import HttpUrl
from .utils import SLOTS
from .validation_context import get_validation_context

AbsolutePathT = TypeVar(
    "AbsolutePathT",
    bound=Union[HttpUrl, AbsoluteDirectory, AbsoluteFilePath, ZipPath],
)


class LightHttpFileDescr(Node):
    """http source with sha256 value (minimal validation)"""

    source: pydantic.HttpUrl
    """file source"""

    sha256: Sha256
    """SHA256 checksum of the source file"""

    def get_reader(
        self,
        *,
        progressbar: Union[Progressbar, Callable[[], Progressbar], bool, None] = None,
    ) -> BytesReader:
        """open the file source (download if needed)"""
        return get_reader(self.source, sha256=self.sha256, progressbar=progressbar)

    download = get_reader
    """alias for get_reader() method"""


class RelativePathBase(RootModel[PurePath], Generic[AbsolutePathT], frozen=True):
    _absolute: AbsolutePathT = PrivateAttr()

    @property
    def path(self) -> PurePath:
        return self.root

    def absolute(  # method not property analog to `pathlib.Path.absolute()`
        self,
    ) -> AbsolutePathT:
        """get the absolute path/url

        (resolved at time of initialization with the root of the ValidationContext)
        """
        return self._absolute

    def model_post_init(self, __context: Any) -> None:
        """set `_absolute` property with validation context at creation time. @private"""
        if self.root.is_absolute():
            raise ValueError(f"{self.root} is an absolute path.")

        if self.root.parts and self.root.parts[0] in ("http:", "https:"):
            raise ValueError(f"{self.root} looks like an http url.")

        self._absolute = (  # pyright: ignore[reportAttributeAccessIssue]
            self.get_absolute(get_validation_context().root)
        )
        super().model_post_init(__context)

    def __str__(self) -> str:
        return self.root.as_posix()

    def __repr__(self) -> str:
        return f"RelativePath('{self}')"

    @model_serializer()
    def format(self) -> str:
        return str(self)

    @abstractmethod
    def get_absolute(
        self, root: Union[RootHttpUrl, AbsoluteDirectory, pydantic.AnyUrl, ZipFile]
    ) -> AbsolutePathT: ...

    def _get_absolute_impl(
        self, root: Union[RootHttpUrl, AbsoluteDirectory, pydantic.AnyUrl, ZipFile]
    ) -> Union[Path, HttpUrl, ZipPath]:
        if isinstance(root, Path):
            return (root / self.root).absolute()

        rel_path = self.root.as_posix().strip("/")
        if isinstance(root, ZipFile):
            return ZipPath(root, rel_path)

        parsed = urlsplit(str(root))
        path = list(parsed.path.strip("/").split("/"))
        if (
            parsed.netloc == "zenodo.org"
            and parsed.path.startswith("/api/records/")
            and parsed.path.endswith("/content")
        ):
            path.insert(-1, rel_path)
        else:
            path.append(rel_path)

        return HttpUrl(
            urlunsplit(
                (
                    parsed.scheme,
                    parsed.netloc,
                    "/".join(path),
                    parsed.query,
                    parsed.fragment,
                )
            )
        )

    @classmethod
    def _validate(cls, value: Union[PurePath, str]):
        if isinstance(value, str) and (
            value.startswith("https://") or value.startswith("http://")
        ):
            raise ValueError(f"{value} looks like a URL, not a relative path")

        return cls(PurePath(value))


class RelativeFilePath(
    RelativePathBase[Union[AbsoluteFilePath, HttpUrl, ZipPath]], frozen=True
):
    """A path relative to the `rdf.yaml` file (also if the RDF source is a URL)."""

    def model_post_init(self, __context: Any) -> None:
        """add validation @private"""
        if not self.root.parts:  # an empty path can only be a directory
            raise ValueError(f"{self.root} is not a valid file path.")

        super().model_post_init(__context)

    def get_absolute(
        self, root: "RootHttpUrl | Path | AnyUrl | ZipFile"
    ) -> "AbsoluteFilePath | HttpUrl | ZipPath":
        absolute = self._get_absolute_impl(root)
        if (
            isinstance(absolute, Path)
            and (context := get_validation_context()).perform_io_checks
            and str(self.root) not in context.known_files
            and not absolute.is_file()
        ):
            raise ValueError(f"{absolute} does not point to an existing file")

        return absolute

    @property
    def suffix(self):
        return self.root.suffix


class RelativeDirectory(
    RelativePathBase[Union[AbsoluteDirectory, HttpUrl, ZipPath]], frozen=True
):
    def get_absolute(
        self, root: "RootHttpUrl | Path | AnyUrl | ZipFile"
    ) -> "AbsoluteDirectory | HttpUrl | ZipPath":
        absolute = self._get_absolute_impl(root)
        if (
            isinstance(absolute, Path)
            and get_validation_context().perform_io_checks
            and not absolute.is_dir()
        ):
            raise ValueError(f"{absolute} does not point to an existing directory")

        return absolute


FileSource = Annotated[
    Union[HttpUrl, RelativeFilePath, FilePath],
    Field(union_mode="left_to_right"),
]
PermissiveFileSource = Union[FileSource, str, pydantic.HttpUrl]


class FileDescr(Node):
    """A file description"""

    source: FileSource
    """File source"""

    sha256: Optional[Sha256] = None
    """SHA256 hash value of the **source** file."""

    @model_validator(mode="after")
    def _validate_sha256(self) -> Self:
        if get_validation_context().perform_io_checks:
            self.validate_sha256()

        return self

    def validate_sha256(self, force_recompute: bool = False) -> None:
        """validate the sha256 hash value of the **source** file"""
        context = get_validation_context()
        src_str = str(self.source)
        if not force_recompute and src_str in context.known_files:
            actual_sha = context.known_files[src_str]
        else:
            reader = get_reader(self.source, sha256=self.sha256)
            if force_recompute:
                actual_sha = get_sha256(reader)
            else:
                actual_sha = reader.sha256

            context.known_files[src_str] = actual_sha

        if actual_sha is None:
            return
        elif self.sha256 == actual_sha:
            pass
        elif self.sha256 is None or context.update_hashes:
            self.sha256 = actual_sha
        elif self.sha256 != actual_sha:
            raise ValueError(
                f"Sha256 mismatch for {self.source}. Expected {self.sha256}, got "
                + f"{actual_sha}. Update expected `sha256` or point to the matching "
                + "file."
            )

    def get_reader(
        self,
        *,
        progressbar: Union[Progressbar, Callable[[], Progressbar], bool, None] = None,
    ):
        """open the file source (download if needed)"""
        return get_reader(self.source, progressbar=progressbar, sha256=self.sha256)

    def download(
        self,
        *,
        progressbar: Union[Progressbar, Callable[[], Progressbar], bool, None] = None,
    ):
        """alias for `.get_reader`"""
        return get_reader(self.source, progressbar=progressbar, sha256=self.sha256)

    @property
    def suffix(self) -> str:
        return self.source.suffix


path_or_url_adapter: "TypeAdapter[Union[FilePath, DirectoryPath, HttpUrl]]" = (
    TypeAdapter(Union[FilePath, DirectoryPath, HttpUrl])
)


@dataclass(frozen=True, **SLOTS)
class WithSuffix:
    suffix: Union[LiteralString, Tuple[LiteralString, ...]]
    case_sensitive: bool

    def __get_pydantic_core_schema__(
        self, source: Type[Any], handler: GetCoreSchemaHandler
    ):
        if not self.suffix:
            raise ValueError("suffix may not be empty")

        schema = handler(source)
        return core_schema.no_info_after_validator_function(
            self.validate,
            schema,
        )

    def validate(
        self, value: Union[FileSource, FileDescr]
    ) -> Union[FileSource, FileDescr]:
        return validate_suffix(value, self.suffix, case_sensitive=self.case_sensitive)


def wo_special_file_name(src: F) -> F:
    if has_valid_bioimageio_yaml_name(src):
        raise ValueError(
            f"'{src}' not allowed here as its filename is reserved to identify"
            + f" '{BIOIMAGEIO_YAML}' (or equivalent) files."
        )

    return src


def has_valid_bioimageio_yaml_name(src: Union[FileSource, FileDescr]) -> bool:
    return is_valid_bioimageio_yaml_name(extract_file_name(src))


def is_valid_bioimageio_yaml_name(file_name: FileName) -> bool:
    for bioimageio_name in ALL_BIOIMAGEIO_YAML_NAMES:
        if file_name == bioimageio_name or file_name.endswith("." + bioimageio_name):
            return True

    return False


def identify_bioimageio_yaml_file_name(file_names: Iterable[FileName]) -> FileName:
    file_names = sorted(file_names)
    for bioimageio_name in ALL_BIOIMAGEIO_YAML_NAMES:
        for file_name in file_names:
            if file_name == bioimageio_name or file_name.endswith(
                "." + bioimageio_name
            ):
                return file_name

    raise ValueError(
        f"No {BIOIMAGEIO_YAML} found in {file_names}. (Looking for '{BIOIMAGEIO_YAML}'"
        + " or or any of the alterntive file names:"
        + f" {ALTERNATIVE_BIOIMAGEIO_YAML_NAMES}, or any file with an extension of"
        + f"  those, e.g. 'anything.{BIOIMAGEIO_YAML}')."
    )


def find_bioimageio_yaml_file_name(path: Union[Path, ZipFile]) -> FileName:
    if isinstance(path, ZipFile):
        file_names = path.namelist()
    elif path.is_file():
        if not zipfile.is_zipfile(path):
            return path.name

        with ZipFile(path, "r") as f:
            file_names = f.namelist()
    else:
        file_names = [p.name for p in path.glob("*")]

    return identify_bioimageio_yaml_file_name(
        file_names
    )  # TODO: try/except with better error message for dir


def ensure_has_valid_bioimageio_yaml_name(src: FileSource) -> FileSource:
    if not has_valid_bioimageio_yaml_name(src):
        raise ValueError(
            f"'{src}' does not have a valid filename to identify"
            + f" '{BIOIMAGEIO_YAML}' (or equivalent) files."
        )

    return src


def ensure_is_valid_bioimageio_yaml_name(file_name: FileName) -> FileName:
    if not is_valid_bioimageio_yaml_name(file_name):
        raise ValueError(
            f"'{file_name}' is not a valid filename to identify"
            + f" '{BIOIMAGEIO_YAML}' (or equivalent) files."
        )

    return file_name


# types as loaded from YAML 1.2 (with ruyaml)
YamlLeafValue = Union[
    bool, _date, _datetime, int, float, str, None
]  # note: order relevant for deserializing
YamlKey = Union[  # YAML Arrays are cast to tuples if used as key in mappings
    YamlLeafValue, Tuple[YamlLeafValue, ...]  # (nesting is not allowed though)
]
if TYPE_CHECKING:
    YamlValue = Union[YamlLeafValue, List["YamlValue"], Dict[YamlKey, "YamlValue"]]
    YamlValueView = Union[
        YamlLeafValue, Sequence["YamlValueView"], Mapping[YamlKey, "YamlValueView"]
    ]
else:
    # for pydantic validation we need to use `TypeAliasType`,
    # see https://docs.pydantic.dev/latest/concepts/types/#named-recursive-types
    # however this results in a partially unknown type with the current pyright 1.1.388
    YamlValue = _TypeAliasType(
        "YamlValue",
        Union[YamlLeafValue, List["YamlValue"], Dict[YamlKey, "YamlValue"]],
    )
    YamlValueView = _TypeAliasType(
        "YamlValueView",
        Union[
            YamlLeafValue,
            Sequence["YamlValueView"],
            Mapping[YamlKey, "YamlValueView"],
        ],
    )

BioimageioYamlContent = Dict[str, YamlValue]
BioimageioYamlContentView = Mapping[str, YamlValueView]
BioimageioYamlSource = Union[
    PermissiveFileSource, ZipFile, BioimageioYamlContent, BioimageioYamlContentView
]


@overload
def deepcopy_yaml_value(value: BioimageioYamlContentView) -> BioimageioYamlContent: ...


@overload
def deepcopy_yaml_value(value: YamlValueView) -> YamlValue: ...


def deepcopy_yaml_value(
    value: Union[BioimageioYamlContentView, YamlValueView],
) -> Union[BioimageioYamlContent, YamlValue]:
    if isinstance(value, str):
        return value
    elif isinstance(value, collections.abc.Mapping):
        return {key: deepcopy_yaml_value(val) for key, val in value.items()}
    elif isinstance(value, collections.abc.Sequence):
        return [deepcopy_yaml_value(val) for val in value]
    else:
        return value


def is_yaml_leaf_value(value: Any) -> TypeGuard[YamlLeafValue]:
    return isinstance(value, (bool, _date, _datetime, int, float, str, type(None)))


def is_yaml_list(value: Any) -> TypeGuard[List[YamlValue]]:
    return is_list(value) and all(is_yaml_value(item) for item in value)


def is_yaml_sequence(value: Any) -> TypeGuard[List[YamlValueView]]:
    return is_sequence(value) and all(is_yaml_value(item) for item in value)


def is_yaml_dict(value: Any) -> TypeGuard[BioimageioYamlContent]:
    return is_dict(value) and all(
        isinstance(key, str) and is_yaml_value(val) for key, val in value.items()
    )


def is_yaml_mapping(value: Any) -> TypeGuard[BioimageioYamlContentView]:
    return is_mapping(value) and all(
        isinstance(key, str) and is_yaml_value_read_only(val)
        for key, val in value.items()
    )


def is_yaml_value(value: Any) -> TypeGuard[YamlValue]:
    return is_yaml_leaf_value(value) or is_yaml_list(value) or is_yaml_dict(value)


def is_yaml_value_read_only(value: Any) -> TypeGuard[YamlValueView]:
    return (
        is_yaml_leaf_value(value) or is_yaml_sequence(value) or is_yaml_mapping(value)
    )


@dataclass(frozen=True, **SLOTS)
class OpenedBioimageioYaml:
    content: BioimageioYamlContent = field(repr=False)
    original_root: Union[AbsoluteDirectory, RootHttpUrl, ZipFile]
    original_source_name: Optional[str]
    original_file_name: FileName
    unparsed_content: str = field(repr=False)


@dataclass(frozen=True, **SLOTS)
class LocalFile:
    path: FilePath
    original_root: Union[AbsoluteDirectory, RootHttpUrl, ZipFile]
    original_file_name: FileName


@dataclass(frozen=True, **SLOTS)
class FileInZip:
    path: ZipPath
    original_root: Union[RootHttpUrl, ZipFile]
    original_file_name: FileName


class HashKwargs(TypedDict):
    sha256: NotRequired[Optional[Sha256]]


_file_source_adapter: TypeAdapter[Union[HttpUrl, RelativeFilePath, FilePath]] = (
    TypeAdapter(FileSource)
)


def interprete_file_source(file_source: PermissiveFileSource) -> FileSource:
    if isinstance(file_source, Path):
        if file_source.is_dir():
            raise FileNotFoundError(
                f"{file_source} is a directory, but expected a file."
            )
        return file_source

    if isinstance(file_source, HttpUrl):
        return file_source

    if isinstance(file_source, pydantic.AnyUrl):
        file_source = str(file_source)

    with get_validation_context().replace(perform_io_checks=False):
        strict = _file_source_adapter.validate_python(file_source)
        if isinstance(strict, Path) and strict.is_dir():
            raise FileNotFoundError(f"{strict} is a directory, but expected a file.")

    return strict


def extract(
    source: Union[FilePath, ZipFile, ZipPath],
    folder: Optional[DirectoryPath] = None,
    overwrite: bool = False,
) -> DirectoryPath:
    extract_member = None
    if isinstance(source, ZipPath):
        extract_member = source.at
        source = source.root

    if isinstance(source, ZipFile):
        zip_context = nullcontext(source)
        if folder is None:
            if source.filename is None:
                folder = Path(mkdtemp())
            else:
                zip_path = Path(source.filename)
                folder = zip_path.with_suffix(zip_path.suffix + ".unzip")
    else:
        zip_context = ZipFile(source, "r")
        if folder is None:
            folder = source.with_suffix(source.suffix + ".unzip")

    if overwrite and folder.exists():
        warnings.warn(f"Overwriting existing unzipped archive at {folder}")

    with zip_context as f:
        if extract_member is not None:
            extracted_file_path = folder / extract_member
            if extracted_file_path.exists() and not overwrite:
                warnings.warn(f"Found unzipped {extracted_file_path}.")
            else:
                _ = f.extract(extract_member, folder)

            return folder

        elif overwrite or not folder.exists():
            f.extractall(folder)
            return folder

        found_content = {p.relative_to(folder).as_posix() for p in folder.glob("*")}
        expected_content = {info.filename for info in f.filelist}
        if expected_missing := expected_content - found_content:
            parts = folder.name.split("_")
            nr, *suffixes = parts[-1].split(".")
            if nr.isdecimal():
                nr = str(int(nr) + 1)
            else:
                nr = f"1.{nr}"

            parts[-1] = ".".join([nr, *suffixes])
            out_path_new = folder.with_name("_".join(parts))
            warnings.warn(
                f"Unzipped archive at {folder} is missing expected files"
                + f" {expected_missing}."
                + f" Unzipping to {out_path_new} instead to avoid overwriting."
            )
            return extract(f, out_path_new, overwrite=overwrite)
        else:
            warnings.warn(
                f"Found unzipped archive with all expected files at {folder}."
            )
            return folder


def get_reader(
    source: Union[PermissiveFileSource, FileDescr, ZipPath],
    /,
    progressbar: Union[Progressbar, Callable[[], Progressbar], bool, None] = None,
    **kwargs: Unpack[HashKwargs],
) -> BytesReader:
    """Open a file `source` (download if needed)"""
    if isinstance(source, FileDescr):
        if "sha256" not in kwargs:
            kwargs["sha256"] = source.sha256

        source = source.source
    elif isinstance(source, str):
        source = interprete_file_source(source)

    if isinstance(source, RelativeFilePath):
        source = source.absolute()
    elif isinstance(source, pydantic.AnyUrl):
        with get_validation_context().replace(perform_io_checks=False):
            source = HttpUrl(source)

    if isinstance(source, HttpUrl):
        return _open_url(source, progressbar=progressbar, **kwargs)

    if isinstance(source, ZipPath):
        if not source.exists():
            raise FileNotFoundError(source)

        f = source.open(mode="rb")
        assert not isinstance(f, TextIOWrapper)
        root = source.root
    elif isinstance(source, Path):
        if source.is_dir():
            raise FileNotFoundError(f"{source} is a directory, not a file")

        if not source.exists():
            raise FileNotFoundError(source)

        f = source.open("rb")
        root = source.parent
    else:
        assert_never(source)

    expected_sha = kwargs.get("sha256")
    if expected_sha is None:
        sha = None
    else:
        sha = get_sha256(f)
        _ = f.seek(0)
        if sha != expected_sha:
            raise ValueError(
                f"SHA256 mismatch for {source}. Expected {expected_sha}, got {sha}."
            )

    return BytesReader(
        f,
        sha256=sha,
        suffix=source.suffix,
        original_file_name=source.name,
        original_root=root,
        is_zipfile=None,
    )


download = get_reader


def _open_url(
    source: HttpUrl,
    /,
    progressbar: Union[Progressbar, Callable[[], Progressbar], bool, None],
    **kwargs: Unpack[HashKwargs],
) -> BytesReader:
    cache = (
        NoopCache[RootHttpUrl](url_hasher=UrlDigest.from_str)
        if get_validation_context().disable_cache
        else settings.disk_cache
    )
    sha = kwargs.get("sha256")
    digest = False if sha is None else ContentDigest.parse(hexdigest=sha)
    source_path = PurePosixPath(
        source.path
        or sha
        or hashlib.sha256(str(source).encode(encoding="utf-8")).hexdigest()
    )

    reader = cache.fetch(
        source,
        fetcher=partial(_fetch_url, progressbar=progressbar),
        force_refetch=digest,
    )
    return BytesReader(
        reader,
        suffix=source_path.suffix,
        sha256=sha,
        original_file_name=source_path.name,
        original_root=source.parent,
        is_zipfile=None,
    )


def _fetch_url(
    source: RootHttpUrl,
    *,
    progressbar: Union[Progressbar, Callable[[], Progressbar], bool, None],
):
    if source.scheme not in ("http", "https"):
        raise NotImplementedError(source.scheme)

    if progressbar is None:
        # chose progressbar option from validation context
        progressbar = get_validation_context().progressbar

    if progressbar is None:
        # default to no progressbar in CI environments
        progressbar = not settings.CI

    if callable(progressbar):
        progressbar = progressbar()

    if isinstance(progressbar, bool) and progressbar:
        progressbar = tqdm(
            ncols=79,
            ascii=bool(sys.platform == "win32"),
            unit="B",
            unit_scale=True,
            leave=True,
        )

    if progressbar is not False:
        progressbar.set_description(f"Downloading {extract_file_name(source)}")

    headers: Dict[str, str] = {}
    if settings.user_agent is not None:
        headers["User-Agent"] = settings.user_agent
    elif settings.CI:
        headers["User-Agent"] = "ci"

    r = httpx.get(
        str(source),
        follow_redirects=True,
        headers=headers,
        timeout=settings.http_timeout,
    )
    _ = r.raise_for_status()

    # set progressbar.total
    total = r.headers.get("content-length")
    if total is not None and not isinstance(total, int):
        try:
            total = int(total)
        except Exception:
            total = None

    if progressbar is not False:
        if total is None:
            progressbar.total = 0
        else:
            progressbar.total = total

    def iter_content():
        for chunk in r.iter_bytes(chunk_size=4096):
            yield chunk
            if progressbar is not False:
                _ = progressbar.update(len(chunk))

        # Make sure the progress bar gets filled even if the actual number
        # is chunks is smaller than expected. This happens when streaming
        # text files that are compressed by the server when sending (gzip).
        # Binary files don't experience this.
        # (adapted from pooch.HttpDownloader)
        if progressbar is not False:
            progressbar.reset()
            if total is not None:
                _ = progressbar.update(total)

            progressbar.close()

    return iter_content()


def extract_file_name(
    src: Union[
        pydantic.HttpUrl, RootHttpUrl, PurePath, RelativeFilePath, ZipPath, FileDescr
    ],
) -> FileName:
    if isinstance(src, FileDescr):
        src = src.source

    if isinstance(src, ZipPath):
        return src.name or src.root.filename or "bioimageio.zip"
    elif isinstance(src, RelativeFilePath):
        return src.path.name
    elif isinstance(src, PurePath):
        return src.name
    else:
        url = urlparse(str(src))
        if (
            url.scheme == "https"
            and url.hostname == "zenodo.org"
            and url.path.startswith("/api/records/")
            and url.path.endswith("/content")
        ):
            return url.path.split("/")[-2]
        else:
            return url.path.split("/")[-1]


def extract_file_descrs(data: YamlValueView):
    collected: List[FileDescr] = []
    with get_validation_context().replace(perform_io_checks=False, log_warnings=False):
        _extract_file_descrs_impl(data, collected)

    return collected


def _extract_file_descrs_impl(data: YamlValueView, collected: List[FileDescr]):
    if isinstance(data, collections.abc.Mapping):
        if "source" in data and "sha256" in data:
            try:
                fd = FileDescr.model_validate(
                    dict(source=data["source"], sha256=data["sha256"])
                )
            except Exception:
                pass
            else:
                collected.append(fd)

        for v in data.values():
            _extract_file_descrs_impl(v, collected)
    elif not isinstance(data, str) and isinstance(data, collections.abc.Sequence):
        for v in data:
            _extract_file_descrs_impl(v, collected)


F = TypeVar("F", bound=Union[FileSource, FileDescr])


def validate_suffix(
    value: F, suffix: Union[str, Sequence[str]], case_sensitive: bool
) -> F:
    """check final suffix"""
    if isinstance(suffix, str):
        suffixes = [suffix]
    else:
        suffixes = suffix

    assert len(suffixes) > 0, "no suffix given"
    assert all(suff.startswith(".") for suff in suffixes), (
        "expected suffixes to start with '.'"
    )
    o_value = value
    if isinstance(value, FileDescr):
        strict = value.source
    else:
        strict = interprete_file_source(value)

    if isinstance(strict, (HttpUrl, AnyUrl)):
        if strict.path is None or "." not in (path := strict.path):
            actual_suffixes = []
        else:
            if (
                strict.host == "zenodo.org"
                and path.startswith("/api/records/")
                and path.endswith("/content")
            ):
                # Zenodo API URLs have a "/content" suffix that should be ignored
                path = path[: -len("/content")]

            actual_suffixes = [f".{path.split('.')[-1]}"]

    elif isinstance(strict, PurePath):
        actual_suffixes = strict.suffixes
    elif isinstance(strict, RelativeFilePath):
        actual_suffixes = strict.path.suffixes
    else:
        assert_never(strict)

    if actual_suffixes:
        actual_suffix = actual_suffixes[-1]
    else:
        actual_suffix = "no suffix"

    if (
        case_sensitive
        and actual_suffix not in suffixes
        or not case_sensitive
        and actual_suffix.lower() not in [s.lower() for s in suffixes]
    ):
        if len(suffixes) == 1:
            raise ValueError(f"Expected suffix {suffixes[0]}, but got {actual_suffix}")
        else:
            raise ValueError(
                f"Expected a suffix from {suffixes}, but got {actual_suffix}"
            )

    return o_value


def populate_cache(sources: Sequence[Union[FileDescr, LightHttpFileDescr]]):
    unique: Set[str] = set()
    for src in sources:
        if src.sha256 is None:
            continue  # not caching without known SHA

        if isinstance(src.source, (HttpUrl, pydantic.AnyUrl)):
            url = str(src.source)
        elif isinstance(src.source, RelativeFilePath):
            if isinstance(absolute := src.source.absolute(), HttpUrl):
                url = str(absolute)
            else:
                continue  # not caching local paths
        elif isinstance(src.source, Path):
            continue  # not caching local paths
        else:
            assert_never(src.source)

        if url in unique:
            continue  # skip duplicate URLs

        unique.add(url)
        _ = src.download()



import io
import shutil
from pathlib import Path

from deepdiff.diff import DeepDiff

from bioimageio.spec.model import v0_5


def test_package(unet2d_path: Path):
    from bioimageio.spec import dump_description, load_description

    descr = load_description(unet2d_path)
    rdf1 = dump_description(descr)
    zip = descr.package()
    descr2 = load_description(zip)
    rdf2 = dump_description(descr2)

    # assert expected differences
    assert (
        rdf1["weights"]["pytorch_state_dict"].pop("source")  # type: ignore
        == "http://files.ilastik.org/bioimageio_test/unet2d_nuclei_broad_weights.torch"
    )
    assert (
        rdf2["weights"]["pytorch_state_dict"].pop("source")  # type: ignore
        == "unet2d_nuclei_broad_weights.torch"
    )

    diff = DeepDiff(rdf1, rdf2)
    assert not diff, diff.pretty()


def test_save_bioimageio_package(unet2d_path: Path):
    from bioimageio.spec import load_model_description, save_bioimageio_package

    package_path = save_bioimageio_package(unet2d_path)
    assert package_path.exists()

    model = load_model_description(package_path)
    assert isinstance(model, v0_5.ModelDescr)


def test_save_bioimageio_package_to_stream(unet2d_path: Path):
    from bioimageio.spec import save_bioimageio_package_to_stream

    output_stream = io.BytesIO()
    stream = save_bioimageio_package_to_stream(unet2d_path, output_stream=output_stream)
    assert isinstance(stream, io.BytesIO)
    assert stream == output_stream


def test_save_bioimageio_package_to_stream_default(unet2d_path: Path):
    from bioimageio.spec import save_bioimageio_package_to_stream

    stream = save_bioimageio_package_to_stream(unet2d_path)
    assert isinstance(stream, io.BytesIO)


def test_save_bioimageio_package_as_folder(unet2d_path: Path, tmp_path: Path):
    from bioimageio.spec import load_description, save_bioimageio_package_as_folder

    package_folder = tmp_path / "package"
    _ = save_bioimageio_package_as_folder(unet2d_path, output_path=package_folder)

    # load package
    model = load_description(package_folder)
    assert isinstance(model, v0_5.ModelDescr)

    # alter package
    doc = model.documentation
    assert isinstance(doc, v0_5.RelativeFilePath)
    new_doc = f"copy_{doc}"
    _ = shutil.move(str(package_folder / str(doc)), str(package_folder / new_doc))
    model.documentation = package_folder / new_doc

    # export altered package
    altered_package = tmp_path / "altered_package"
    altered_package = save_bioimageio_package_as_folder(
        model, output_path=altered_package, weights_priority_order=["onnx"]
    )
    assert (altered_package / new_doc).exists(), altered_package / new_doc

    # load altered package
    reloaded_model = load_description(altered_package)
    assert isinstance(reloaded_model, v0_5.ModelDescr)
    assert str(reloaded_model.documentation).startswith("copy_")



from abc import ABC, abstractmethod
from typing import Any, Dict, Generic, Type, TypeVar

from pydantic import SerializationInfo, SerializerFunctionWrapHandler, model_serializer

from .node import Node
from .validated_string import ValidatedString

InnerNodeT = TypeVar("InnerNodeT", bound=Node)


class ValidatedStringWithInnerNode(ABC, ValidatedString, Generic[InnerNodeT]):
    """A validated string with further validation and serialization using a `Node`"""

    _inner_node_class: Type[InnerNodeT]
    _inner_node: InnerNodeT  # initalized in _after_validator called in __new__

    @model_serializer(mode="wrap")
    def _serialize(
        self, handler: SerializerFunctionWrapHandler, info: SerializationInfo
    ):
        _ = self._inner_node.model_dump(mode=info.mode)
        return handler(
            self,
            info,  # pyright: ignore[reportArgumentType]  # taken from pydantic docs
        )

    @classmethod
    @abstractmethod
    def _get_data(cls, valid_string_data: str) -> Dict[str, Any]: ...

    def _after_validator(self):
        data = self._get_data(self._validated)
        self._inner_node = self._inner_node_class.model_validate(data)
        return super()._after_validator()



from __future__ import annotations

import collections.abc
import warnings
from typing import (
    Any,
    Literal,
    Mapping,
    Optional,
    Type,
    Union,
)

import pydantic
from typing_extensions import Self

from .type_guards import is_kwargs
from .validation_context import ValidationContext, get_validation_context


def _node_title_generator(node: Type[Node]) -> str:
    return (
        f"{node.implemented_type} {node.implemented_format_version}"  # pyright: ignore[reportAttributeAccessIssue]
        if hasattr(node, "implemented_type")
        and hasattr(node, "implemented_format_version")
        else f"{node.__module__.replace('bioimageio.spec.', '')}.{node.__name__}"
    )


class Node(
    pydantic.BaseModel,
    allow_inf_nan=False,
    extra="forbid",
    frozen=False,
    model_title_generator=_node_title_generator,
    populate_by_name=True,
    revalidate_instances="always",
    use_attribute_docstrings=True,
    validate_assignment=True,
    validate_default=True,
    validate_return=True,
):
    """"""  # empty docstring to remove all pydantic docstrings from the pdoc spec docs

    @classmethod
    def model_validate(
        cls,
        obj: Union[Any, Mapping[str, Any]],
        *,
        strict: Optional[bool] = None,
        extra: Optional[Literal["allow", "ignore", "forbid"]] = None,
        from_attributes: Optional[bool] = None,
        context: Union[ValidationContext, Mapping[str, Any], None] = None,
        by_alias: Optional[bool] = None,
        by_name: Optional[bool] = None,
    ) -> Self:
        """Validate a pydantic model instance.

        Args:
            obj: The object to validate.
            strict: Whether to raise an exception on invalid fields.
            from_attributes: Whether to extract data from object attributes.
            context: Additional context to pass to the validator.

        Raises:
            ValidationError: If the object failed validation.

        Returns:
            The validated description instance.
        """
        __tracebackhide__ = True

        if context is None:
            context = get_validation_context()
        elif isinstance(context, collections.abc.Mapping):
            context = ValidationContext(**context)

        assert not isinstance(obj, collections.abc.Mapping) or is_kwargs(obj), obj

        # TODO: pass on extra with pydantic >=2.12
        if extra is not None:
            warnings.warn("`extra` argument is currently ignored")

        with context:
            # use validation context as context manager for equal behavior of __init__ and model_validate
            return super().model_validate(
                obj, strict=strict, from_attributes=from_attributes
            )



def test_license_zenodo():
    """test that all licenses are known or not known by zenodo.
    Run scripts/update_spdx_licenses_zenodo.py to fix this test"""
    from bioimageio.spec.utils import get_spdx_licenses

    for lic in get_spdx_licenses()["licenses"]:
        assert isinstance(lic["isKnownByZenodo"], bool), lic["licenseId"]



import collections.abc
import warnings
from functools import partial
from pathlib import PurePosixPath
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

import numpy as np
from imageio.v3 import imwrite  # pyright: ignore[reportUnknownVariableType]
from loguru import logger
from numpy.typing import NDArray
from typing_extensions import assert_never

from bioimageio.spec._internal.validation_context import get_validation_context
from bioimageio.spec.model.v0_5 import (
    IntervalOrRatioDataDescr,
    KerasHdf5WeightsDescr,
    NominalOrOrdinalDataDescr,
    OnnxWeightsDescr,
    PytorchStateDictWeightsDescr,
    TensorflowJsWeightsDescr,
    TensorflowSavedModelBundleWeightsDescr,
    TensorId,
    TorchscriptWeightsDescr,
)

from ._internal.io import RelativeFilePath, get_reader
from ._internal.io_utils import load_array
from ._version import VERSION
from .model import ModelDescr
from .utils import get_spdx_licenses, load_image

HF_KNOWN_LICENSES = (
    "apache-2.0",
    "mit",
    "openrail",
    "bigscience-openrail-m",
    "creativeml-openrail-m",
    "bigscience-bloom-rail-1.0",
    "bigcode-openrail-m",
    "afl-3.0",
    "artistic-2.0",
    "bsl-1.0",
    "bsd",
    "bsd-2-clause",
    "bsd-3-clause",
    "bsd-3-clause-clear",
    "c-uda",
    "cc",
    "cc0-1.0",
    "cc-by-2.0",
    "cc-by-2.5",
    "cc-by-3.0",
    "cc-by-4.0",
    "cc-by-sa-3.0",
    "cc-by-sa-4.0",
    "cc-by-nc-2.0",
    "cc-by-nc-3.0",
    "cc-by-nc-4.0",
    "cc-by-nd-4.0",
    "cc-by-nc-nd-3.0",
    "cc-by-nc-nd-4.0",
    "cc-by-nc-sa-2.0",
    "cc-by-nc-sa-3.0",
    "cc-by-nc-sa-4.0",
    "cdla-sharing-1.0",
    "cdla-permissive-1.0",
    "cdla-permissive-2.0",
    "wtfpl",
    "ecl-2.0",
    "epl-1.0",
    "epl-2.0",
    "etalab-2.0",
    "eupl-1.1",
    "eupl-1.2",
    "agpl-3.0",
    "gfdl",
    "gpl",
    "gpl-2.0",
    "gpl-3.0",
    "lgpl",
    "lgpl-2.1",
    "lgpl-3.0",
    "isc",
    "h-research",
    "intel-research",
    "lppl-1.3c",
    "ms-pl",
    "apple-ascl",
    "apple-amlr",
    "mpl-2.0",
    "odc-by",
    "odbl",
    "openmdw-1.0",
    "openrail++",
    "osl-3.0",
    "postgresql",
    "ofl-1.1",
    "ncsa",
    "unlicense",
    "zlib",
    "pddl",
    "lgpl-lr",
    "deepfloyd-if-license",
    "fair-noncommercial-research-license",
    "llama2",
    "llama3",
    "llama3.1",
    "llama3.2",
    "llama3.3",
    "llama4",
    "grok2-community",
    "gemma",
)


def _generate_png_from_tensor(tensor: NDArray[np.generic]) -> Optional[bytes]:
    """Generate PNG bytes from a sample tensor.

    Prefers 2D slices from multi-dimensional arrays.
    Returns PNG bytes or None if generation fails.
    """
    try:
        # Squeeze out singleton dimensions
        arr = np.squeeze(tensor)

        # Handle different dimensionalities
        if arr.ndim == 2:
            img_data = arr
        elif arr.ndim == 3:
            # Could be (H, W, C) or (Z, H, W)
            if arr.shape[-1] in [1, 3, 4]:  # Likely channels last
                img_data = arr
            else:  # Take middle slice
                img_data = arr[arr.shape[0] // 2]
        elif arr.ndim == 4:
            # Take middle slices (e.g., batch, z, y, x)
            img_data = (
                arr[0, arr.shape[1] // 2]
                if arr.shape[0] == 1
                else arr[arr.shape[0] // 2, arr.shape[1] // 2]
            )
        elif arr.ndim > 4:
            # Take middle slices of all extra dimensions
            slices = tuple(s // 2 for s in arr.shape[:-2])
            img_data = arr[slices]
        else:
            return None

        # Normalize to 0-255 uint8
        img_data = np.squeeze(img_data)
        if img_data.dtype != np.uint8:
            img_min, img_max = img_data.min(), img_data.max()
            if img_max > img_min:
                img_data: NDArray[Any] = (img_data - img_min) / (img_max - img_min)
            else:
                img_data = np.zeros_like(img_data)
            img_data = (img_data * 255).astype(np.uint8)
        return imwrite("<bytes>", img_data, extension=".png")
    except Exception:
        return None


def _get_io_description(
    model: ModelDescr,
) -> Tuple[str, Dict[str, bytes], List[TensorId], List[TensorId]]:
    """Generate a description of model inputs and outputs with sample images.

    Returns:
        A tuple of (markdown_string, referenced_files_dict, input_ids, output_ids) where referenced_files_dict maps
        filenames to file bytes.
    """
    markdown_string = ""
    referenced_files: dict[str, bytes] = {}
    input_ids: List[TensorId] = []
    output_ids: List[TensorId] = []

    def format_data_descr(
        d: Union[
            NominalOrOrdinalDataDescr,
            IntervalOrRatioDataDescr,
            Sequence[Union[NominalOrOrdinalDataDescr, IntervalOrRatioDataDescr]],
        ],
    ) -> str:
        ret = ""
        if isinstance(d, NominalOrOrdinalDataDescr):
            ret += f"  - Values: {d.values}\n"
        elif isinstance(d, IntervalOrRatioDataDescr):
            ret += f"  - Value unit: {d.unit}\n"
            ret += f"  - Value scale factor: {d.scale}\n"
            if d.offset is not None:
                ret += f"  - Value offset: {d.offset}\n"
            elif d.range[0] is not None:
                ret += f"  - Value minimum: {d.range[0]}\n"
            elif d.range[1] is not None:
                ret += f"  - Value maximum: {d.range[1]}\n"
        elif isinstance(d, collections.abc.Sequence):
            for dd in d:
                ret += format_data_descr(dd)
        else:
            assert_never(d)

        return ret

    # Input descriptions
    if model.inputs:
        markdown_string += "\n- **Input specifications:**\n"

        for inp in model.inputs:
            input_ids.append(inp.id)
            axes_str = ", ".join(str(a.id) for a in inp.axes)
            shape_str = " × ".join(
                str(a.size) if isinstance(a.size, int) else str(a.size)
                for a in inp.axes
            )

            markdown_string += f"  `{inp.id}`: {inp.description or ''}\n\n"
            markdown_string += f"  - Axes: `{axes_str}`\n"
            markdown_string += f"  - Shape: `{shape_str}`\n"
            markdown_string += f"  - Data type: `{inp.dtype}`\n"
            markdown_string += format_data_descr(inp.data)

            # Try to load and display sample_tensor (preferred) or test_tensor
            img_bytes = None
            if inp.sample_tensor is not None:
                try:
                    arr = load_image(inp.sample_tensor)
                    img_bytes = _generate_png_from_tensor(arr)
                except Exception as e:
                    logger.error("failed to generate input sample image: {}", e)

            if img_bytes is None and inp.test_tensor is not None:
                try:
                    arr = load_array(inp.test_tensor)
                    img_bytes = _generate_png_from_tensor(arr)
                except Exception as e:
                    logger.error(
                        "failed to generate input sample image from test data: {}", e
                    )

            if img_bytes:
                filename = f"images/input_{inp.id}_sample.png"
                referenced_files[filename] = img_bytes
                markdown_string += f"  - example\n    ![{inp.id} sample]({filename})\n"

    # Output descriptions
    if model.outputs:
        markdown_string += "\n- **Output specifications:**\n"
        for out in model.outputs:
            output_ids.append(out.id)
            axes_str = ", ".join(str(a.id) for a in out.axes)
            shape_str = " × ".join(
                str(a.size) if isinstance(a.size, int) else str(a.size)
                for a in out.axes
            )

            markdown_string += f"  `{out.id}`: {out.description or ''}\n"
            markdown_string += f"  - Axes: `{axes_str}`\n"
            markdown_string += f"  - Shape: `{shape_str}`\n"
            markdown_string += f"  - Data type: `{out.dtype}`\n"
            markdown_string += format_data_descr(out.data)

            # Try to load and display sample_tensor (preferred) or test_tensor
            img_bytes = None
            if out.sample_tensor is not None:
                try:
                    arr = load_image(out.sample_tensor)
                    img_bytes = _generate_png_from_tensor(arr)
                except Exception as e:
                    logger.error("failed to generate output sample image: {}", e)

            if img_bytes is None and out.test_tensor is not None:
                try:
                    arr = load_array(out.test_tensor)
                    img_bytes = _generate_png_from_tensor(arr)
                except Exception as e:
                    logger.error(
                        "failed to generate output sample image from test data: {}", e
                    )

            if img_bytes:
                filename = f"images/output_{out.id}_sample.png"
                referenced_files[filename] = img_bytes
                markdown_string += f"  - example\n    {out.id} sample]({filename})\n"

    return markdown_string, referenced_files, input_ids, output_ids


def create_huggingface_model_card(
    model: ModelDescr, *, repo_id: str
) -> Tuple[str, Dict[str, bytes]]:
    """Create a Hugging Face model card for a BioImage.IO model.

    Returns:
        A tuple of (markdown_string, images_dict) where images_dict maps
        filenames to PNG bytes that should be saved alongside the markdown.
    """
    if model.version is None:
        model_version = ""
    else:
        model_version = f"\n- **model version:** {model.version}"

    if model.documentation is None:
        additional_model_doc = ""
    else:
        doc_reader = get_reader(model.documentation)
        local_doc_path = f"package/{doc_reader.original_file_name}"
        model = model.model_copy()
        with get_validation_context().replace(perform_io_checks=False):
            model.documentation = RelativeFilePath(PurePosixPath(local_doc_path))

        additional_model_doc = f"\n- **Additional model documentation:** [{local_doc_path}]({local_doc_path})"

    if model.cite:
        developed_by = "\n- **Developed by:** " + (
            "".join(
                (
                    f"\n    - {c.text}: "
                    + (f"https://www.doi.org/{c.doi}" if c.doi else str(c.url))
                )
                for c in model.cite
            )
        )
    else:
        developed_by = ""

    if model.config.bioimageio.funded_by:
        funded_by = f"\n- **Funded by:** {model.config.bioimageio.funded_by}"
    else:
        funded_by = ""

    if model.authors:
        shared_by = "\n- **Shared by:** " + (
            "".join(
                (
                    f"\n    - {a.name}"
                    + (f", {a.affiliation}" if a.affiliation else "")
                    + (
                        f", [https://orcid.org/{a.orcid}](https://orcid.org/{a.orcid})"
                        if a.orcid
                        else ""
                    )
                    + (
                        f", [https://github.com/{a.github_user}](https://github.com/{a.github_user})"
                        if a.github_user
                        else ""
                    )
                    for a in model.authors
                )
            )
        )
    else:
        shared_by = ""

    if model.config.bioimageio.architecture_type:
        model_type = f"\n- **Model type:** {model.config.bioimageio.architecture_type}"
    else:
        model_type = ""

    if model.config.bioimageio.modality:
        model_modality = f"\n- **Modality:** {model.config.bioimageio.modality}"
    else:
        model_modality = ""

    if model.config.bioimageio.target_structure:
        target_structures = "\n- **Target structures:** " + ", ".join(
            model.config.bioimageio.target_structure
        )
    else:
        target_structures = ""

    if model.config.bioimageio.task:
        task_type = f"\n- **Task type:** {model.config.bioimageio.task}"
    else:
        task_type = ""

    if model.parent:
        finetuned_from = f"\n- **Finetuned from model:** {model.parent.id}"
    else:
        finetuned_from = ""

    repository = (
        f"[{model.git_repo}]({model.git_repo})" if model.git_repo else "missing"
    )

    dl_framework_parts: List[str] = []
    training_frameworks: List[str] = []
    model_size: Optional[str] = None
    for weights in model.weights.available_formats.values():
        if isinstance(weights, (PytorchStateDictWeightsDescr, TorchscriptWeightsDescr)):
            dl_framework_version = weights.pytorch_version
        elif isinstance(
            weights,
            (
                TensorflowSavedModelBundleWeightsDescr,
                TensorflowJsWeightsDescr,
                KerasHdf5WeightsDescr,
            ),
        ):
            dl_framework_version = weights.tensorflow_version
        elif isinstance(weights, OnnxWeightsDescr):
            dl_framework_version = f"opset version: {weights.opset_version}"
        else:
            assert_never(weights)

        if weights.parent is None:
            training_frameworks.append(weights.weights_format_name)

        dl_framework_parts.append(
            f"\n  - {weights.weights_format_name}: {dl_framework_version}"
        )

        if model_size is None:
            s = 0
            r = weights.get_reader()
            for chunk in iter(partial(r.read, 128 * 1024), b""):
                s += len(chunk)

            if model.config.bioimageio.model_parameter_count is not None:
                if model.config.bioimageio.model_parameter_count < 1e9:
                    model_size = f"{model.config.bioimageio.model_parameter_count / 1e6:.2f} million parameters, "
                else:
                    model_size = f"{model.config.bioimageio.model_parameter_count / 1e9:.2f} billion parameters, "
            else:
                model_size = ""

            if s < 1e9:
                model_size += f"{s / 1e6:.2f} MB"
            else:
                model_size += f"{s / 1e9:.2f} GB"

    dl_frameworks = "".join(dl_framework_parts)
    if len(training_frameworks) > 1:
        warnings.warn(
            "Multiple training frameworks detected. (Some weight formats are probably missing a `parent` reference.)"
        )

    if (
        model.weights.pytorch_state_dict is not None
        and model.weights.pytorch_state_dict.dependencies is not None
    ):
        env_reader = model.weights.pytorch_state_dict.dependencies.get_reader()
        dependencies = f"Dependencies for Pytorch State dict weights are listed in [{env_reader.original_file_name}](package/{env_reader.original_file_name})."
    else:
        dependencies = "None beyond the respective framework library."

    out_of_scope_use = (
        model.config.bioimageio.out_of_scope_use
        if model.config.bioimageio.out_of_scope_use
        else """missing; therefore these typical limitations should be considered:

- *Likely not suitable for diagnostic purposes.*
- *Likely not validated for different imaging modalities than present in the training data.*
- *Should not be used without proper validation on user's specific datasets.*

"""
    )

    environmental_impact = model.config.bioimageio.environmental_impact.format_md()
    if environmental_impact:
        environmental_impact_toc_entry = (
            "\n- [Environmental Impact](#environmental-impact)"
        )
    else:
        environmental_impact_toc_entry = ""

    evaluation_parts: List[str] = []
    n_evals = 0
    for e in model.config.bioimageio.evaluations:
        if e.dataset_role == "independent":
            continue  # treated separately below

        n_evals += 1
        n_evals_str = "" if n_evals == 1 else f" {n_evals}"
        evaluation_parts.append(f"\n# Evaluation{n_evals_str}\n")
        evaluation_parts.append(e.format_md())

    n_evals = 0
    for e in model.config.bioimageio.evaluations:
        if e.dataset_role != "independent":
            continue  # treated separately above

        n_evals += 1
        n_evals_str = "" if n_evals == 1 else f" {n_evals}"

        evaluation_parts.append(f"### Validation on External Data{n_evals_str}\n")
        evaluation_parts.append(e.format_md())

    if evaluation_parts:
        evaluation = "\n".join(evaluation_parts)
        evaluation_toc_entry = "\n- [Evaluation](#evaluation)"
    else:
        evaluation = ""
        evaluation_toc_entry = ""

    training_details = ""
    if model.config.bioimageio.training.training_preprocessing:
        training_details += f"### Preprocessing\n\n{model.config.bioimageio.training.training_preprocessing}\n\n"

    training_details += "### Training Hyperparameters\n\n"
    training_details += f"- **Framework:** {' / '.join(training_frameworks)}"
    if model.config.bioimageio.training.training_epochs is not None:
        training_details += (
            f"- **Epochs:** {model.config.bioimageio.training.training_epochs}\n"
        )

    if model.config.bioimageio.training.training_batch_size is not None:
        training_details += f"- **Batch size:** {model.config.bioimageio.training.training_batch_size}\n"

    if model.config.bioimageio.training.initial_learning_rate is not None:
        training_details += f"- **Initial learning rate:** {model.config.bioimageio.training.initial_learning_rate}\n"

    if model.config.bioimageio.training.learning_rate_schedule is not None:
        training_details += f"- **Learning rate schedule:** {model.config.bioimageio.training.learning_rate_schedule}\n"

    if model.config.bioimageio.training.loss_function is not None:
        training_details += (
            f"- **Loss function:** {model.config.bioimageio.training.loss_function}"
        )
        if model.config.bioimageio.training.loss_function_kwargs:
            training_details += (
                f" with {model.config.bioimageio.training.loss_function_kwargs}"
            )
        training_details += "\n"

    if model.config.bioimageio.training.optimizer is not None:
        training_details += (
            f"- **Optimizer:** {model.config.bioimageio.training.optimizer}"
        )
        if model.config.bioimageio.training.optimizer_kwargs:
            training_details += (
                f" with {model.config.bioimageio.training.optimizer_kwargs}"
            )
        training_details += "\n"

    if model.config.bioimageio.training.regularization is not None:
        training_details += (
            f"- **Regularization:** {model.config.bioimageio.training.regularization}\n"
        )

    speeds_sizes_times = "### Speeds, Sizes, Times\n\n"
    if model.config.bioimageio.training.training_duration is not None:
        speeds_sizes_times += f"- **Training time:** {'{:.2f}'.format(model.config.bioimageio.training.training_duration)}\n"

    speeds_sizes_times += f"- **Model size:** {model_size}\n"
    if model.config.bioimageio.inference_time:
        speeds_sizes_times += (
            f"- **Inference time:** {model.config.bioimageio.inference_time}\n"
        )

    if model.config.bioimageio.memory_requirements_inference:
        speeds_sizes_times += f"- **Memory requirements:** {model.config.bioimageio.memory_requirements_inference}\n"

    model_arch_and_objective = "## Model Architecture and Objective\n\n"
    if (
        model.config.bioimageio.architecture_type
        or model.config.bioimageio.architecture_description
    ):
        model_arch_and_objective += (
            f"- **Architecture:** {model.config.bioimageio.architecture_type or ''}"
            + (
                " --- "
                if model.config.bioimageio.architecture_type
                and model.config.bioimageio.architecture_description
                else ""
            )
            + (
                model.config.bioimageio.architecture_description
                if model.config.bioimageio.architecture_description is not None
                else ""
            )
            + "\n"
        )

    io_desc, referenced_files, input_ids, output_ids = _get_io_description(model)
    predict_snippet_inputs = str(
        {input_id: "<path or tensor>" for input_id in input_ids}
    )
    model_arch_and_objective += io_desc

    hardware_requirements = "\n### Hardware Requirements\n"
    if model.config.bioimageio.memory_requirements_training is not None:
        hardware_requirements += f"- **Training:** GPU memory: {model.config.bioimageio.memory_requirements_training}\n"

    if model.config.bioimageio.memory_requirements_inference is not None:
        hardware_requirements += f"- **Inference:** GPU memory: {model.config.bioimageio.memory_requirements_inference}\n"

    hardware_requirements += f"- **Storage:** Model size: {model_size}\n"

    if model.license is None:
        license = "unknown"
        license_meta = "unknown"
    else:
        spdx_licenses = get_spdx_licenses()
        matches = [
            (entry["name"], entry["reference"])
            for entry in spdx_licenses["licenses"]
            if entry["licenseId"].lower() == model.license.lower()
        ]
        if matches:
            if len(matches) > 1:
                logger.warning(
                    "Multiple SPDX license matches found for '{}', using the first one.",
                    model.license,
                )
            name, reference = matches[0]
            license = f"[{name}]({reference})"
            if model.license.lower() in HF_KNOWN_LICENSES:
                license_meta = model.license.lower()
            else:
                license_meta = f"other\nlicense_name: {model.license.lower()}\nlicense_link: {reference}"
        else:
            if model.license.lower() in HF_KNOWN_LICENSES:
                license_meta = model.license.lower()
            else:
                license_meta = "unknown"

            license = model.license.lower()

    base_model = (
        f"\nbase_model: {model.parent.id[len('huggingface/') :]}"
        if model.parent is not None and model.parent.id.startswith("huggingface/")
        else ""
    )
    dataset_meta = (
        f"\ndataset: {model.training_data.id[len('huggingface/') :]}"
        if model.training_data is not None
        and model.training_data.id is not None
        and model.training_data.id.startswith("huggingface/")
        else ""
    )
    if model.covers:
        cover_image_reader = get_reader(model.covers[0])
        cover_image_bytes = cover_image_reader.read()
        cover_image_filename = f"images/{cover_image_reader.original_file_name}"
        referenced_files[cover_image_filename] = cover_image_bytes
        cover_image_md = f"\n![cover image]({cover_image_filename})\n\n"
        thumbnail_meta = (
            f"\nthumbnail: {cover_image_filename}"  # TODO: fix this to be a proper URL
        )

    else:
        cover_image_md = ""
        thumbnail_meta = ""

    # TODO: add pipeline_tag to metadata
    readme = f"""---
license: {license_meta}{thumbnail_meta}
tags: {list({"biology"}.union(set(model.tags)))}
language: [en]
library_name: bioimageio{base_model}{dataset_meta}
---
# {model.name}{cover_image_md}

{model.description or ""}


# Table of Contents

- [Model Details](#model-details)
- [Uses](#uses)
- [Bias, Risks, and Limitations](#bias-risks-and-limitations)
- [How to Get Started with the Model](#how-to-get-started-with-the-model)
- [Training Details](#training-details){evaluation_toc_entry}{
        environmental_impact_toc_entry
    }
- [Technical Specifications](#technical-specifications)


# Model Details

## Model Description
{model_version}{additional_model_doc}{developed_by}{funded_by}{shared_by}{model_type}{
        model_modality
    }{target_structures}{task_type}
- **License:** {license}{finetuned_from}

## Model Sources

- **Repository:** {repository}
- **Paper:** see [**Developed by**](#model-description)

# Uses

## Direct Use

This model is compatible with the bioimageio.spec Python package (version >= {
        VERSION
    }) and the bioimageio.core Python package supporting model inference in Python code or via the `bioimageio` CLI.

```python
from bioimageio.core import predict

output_sample = predict(
    "huggingface/{repo_id}/{model.version or "draft"}",
    inputs={predict_snippet_inputs},
)

output_tensor = output_sample.members["{
        output_ids[0] if output_ids else "<output_id>"
    }"]
xarray_dataarray = output_tensor.data
numpy_ndarray = output_tensor.data.to_numpy()
```

## Downstream Use

Specific bioimage.io partner tool compatibilities may be reported at [Compatibility Reports](https://bioimage-io.github.io/collection/latest/compatibility/#compatibility-by-resource).
{
        "Training (and fine-tuning) code may be available at " + model.git_repo + "."
        if model.git_repo
        else ""
    }

## Out-of-Scope Use

{out_of_scope_use}


{model.config.bioimageio.bias_risks_limitations.format_md()}

# How to Get Started with the Model

You can use "huggingface/{repo_id}/{
        model.version or "draft"
    }" as the resource identifier to load this model directly from the Hugging Face Hub using bioimageio.spec or bioimageio.core.

See [bioimageio.core documentation: Get started](https://bioimage-io.github.io/core-bioimage-io-python/latest/get-started) for instructions on how to load and run this model using the `bioimageio.core` Python package or the bioimageio CLI.

# Training Details

## Training Data

{
        "This model was trained on `" + str(model.training_data.id) + "`."
        if model.training_data is not None
        else "missing"
    }

## Training Procedure

{training_details}

{speeds_sizes_times}
{evaluation}
{environmental_impact}

# Technical Specifications

{model_arch_and_objective}

## Compute Infrastructure

{hardware_requirements}

### Software

- **Framework:** {dl_frameworks}
- **Libraries:** {dependencies}
- **BioImage.IO partner compatibility:** [Compatibility Reports](https://bioimage-io.github.io/collection/latest/compatibility/#compatibility-by-resource)

---

*This model card was created using the template of the bioimageio.spec Python Package, which intern is based on the BioImage Model Zoo template, incorporating best practices from the Hugging Face Model Card Template. For more information on contributing models, visit [bioimage.io](https://bioimage.io).*

---

**References:**

- [Hugging Face Model Card Template](https://huggingface.co/docs/hub/en/model-card-annotated)
- [Hugging Face modelcard_template.md](https://github.com/huggingface/huggingface_hub/blob/b9decfdf9b9a162012bc52f260fd64fc37db660e/src/huggingface_hub/templates/modelcard_template.md)
- [BioImage Model Zoo Documentation](https://bioimage.io/docs/)
- [Model Cards for Model Reporting](https://arxiv.org/abs/1810.03993)
- [bioimageio.spec Python Package](https://bioimage-io.github.io/spec-bioimage-io)
"""

    return readme, referenced_files
